from unittest.mock import patch

import factory
import pytest
from django.contrib.contenttypes.models import ContentType
from wbcore.contrib.authentication.factories import GroupFactory, UserFactory
from wbcore.contrib.authentication.models import Permission
from wbcore.contrib.directory.factories import CompanyFactory, PersonFactory
from wbcore.contrib.directory.models import Person
from wbcore.contrib.workflow.models import (
    Condition,
    DataValue,
    Process,
    ProcessStep,
    Transition,
    Workflow,
)
from wbcore.contrib.workflow.models.workflow import re_render_workflow_graph
from wbcore.contrib.workflow.sites import workflow_site
from wbcore.test.utils import get_or_create_superuser


@pytest.mark.django_db
class TestWorkflow:
    @patch("wbcore.contrib.workflow.models.workflow.transaction.on_commit")
    def test_start_workflow_without_attached_instance(
        self, mock_commit, workflow_factory, data_factory, start_step_factory
    ):
        workflow = workflow_factory(model=None, status_field=None, preserve_instance=False)
        data1 = data_factory(workflow=workflow)
        default_value = "Test"
        data2 = data_factory(workflow=workflow, default=default_value)
        start_step = start_step_factory(workflow=workflow)
        workflow.start_workflow(start_step)
        process_qs = Process.objects.filter(workflow=workflow, instance_id=None, content_type=None)
        assert process_qs.exists()
        assert DataValue.objects.filter(data=data1, process=process_qs.first()).exists()
        assert DataValue.objects.filter(data=data2, process=process_qs.first(), value=default_value).exists()
        assert mock_commit.called

    @patch("wbcore.contrib.workflow.models.workflow.transaction.on_commit")
    def test_start_workflow_with_attached_instance(
        self, mock_commit, workflow_factory, start_step_factory, data_factory
    ):
        attached_person = PersonFactory()
        workflow = workflow_factory(preserve_instance=True)
        data1 = data_factory(workflow=workflow)
        default_value = "Test"
        data2 = data_factory(workflow=workflow, default=default_value)
        start_step = start_step_factory()
        workflow_site.registered_model_classes_serializer_map[Person] = (
            "wbcore.contrib.directory.serializers.PersonModelSerializer"
        )
        workflow.start_workflow(start_step, attached_person)
        process_qs = Process.objects.filter(
            workflow=workflow,
            instance_id=attached_person.pk,
            content_type=ContentType.objects.get_for_model(Person),
            preserved_instance__isnull=False,
        )
        assert process_qs.exists()
        assert DataValue.objects.filter(data=data1, process=process_qs.first()).exists()
        assert DataValue.objects.filter(data=data2, process=process_qs.first(), value=default_value).exists()
        assert mock_commit.called

    def test_get_start_steps_for_workflow(self, workflow_factory, start_step_factory):
        workflow = workflow_factory(model=None, single_instance_execution=False, status_field=None)
        start_step = start_step_factory(workflow=workflow)
        start_step_factory()
        start_transitions = workflow.get_start_steps_for_workflow()
        assert list(start_transitions) == [start_step]

    def test_get_start_steps_for_workflow_single_instance_execution(
        self, workflow_factory, process_factory, start_step_factory
    ):
        workflow = workflow_factory(model=None, single_instance_execution=True, status_field=None)
        start_step_factory(workflow=workflow)
        start_step_factory()
        process_factory(state=Process.ProcessState.ACTIVE, workflow=workflow)
        assert not workflow.get_start_steps_for_workflow().exists()

    def test_get_start_steps_for_workflow_single_instance_execution_no_running_process(
        self, workflow_factory, process_factory, start_step_factory
    ):
        workflow = workflow_factory(model=None, single_instance_execution=True, status_field=None)
        start_step = start_step_factory(workflow=workflow)
        start_step_factory()
        process_factory(state=Process.ProcessState.FINISHED, workflow=workflow)
        start_transitions = workflow.get_start_steps_for_workflow()
        assert list(start_transitions) == [start_step]

    def test_get_start_steps_for_workflow_with_attached_instance(self, workflow_factory, start_step_factory):
        workflow = workflow_factory()
        start_step_factory(workflow=workflow)
        start_step_factory()
        assert not workflow.get_start_steps_for_workflow().exists()

    def test_get_start_steps_for_instance(self, workflow_factory, start_step_factory):
        person = PersonFactory()
        workflow = workflow_factory()
        start_step = start_step_factory(workflow=workflow, status=person.first_name)
        start_step_factory(workflow=workflow)
        assert Workflow.get_start_steps_for_instance(person) == [start_step]

    def test_get_start_steps_for_instance_no_model_attached(self, workflow_factory, start_step_factory):
        person = PersonFactory()
        workflow = workflow_factory(model=None, status_field=None)
        start_step_factory(workflow=workflow, status=person.first_name)
        assert not Workflow.get_start_steps_for_instance(person)

    def test_get_start_steps_for_instance_wrong_instance(self, workflow_factory, start_step_factory):
        company = CompanyFactory()
        workflow = workflow_factory()
        start_step_factory(workflow=workflow)
        assert not Workflow.get_start_steps_for_instance(company)

    def test_get_start_steps_for_instance_single_instance_execution(
        self, workflow_factory, process_factory, start_step_factory
    ):
        person = PersonFactory()
        workflow = workflow_factory(single_instance_execution=True)
        start_step_factory(workflow=workflow, status=person.first_name)
        process_factory(finished=None, workflow=workflow, instance=person)
        assert not Workflow.get_start_steps_for_instance(person)

    def test_get_start_steps_for_instance_single_instance_execution_no_running_process(
        self, workflow_factory, process_factory, start_step_factory
    ):
        attached_instance = PersonFactory()
        workflow = workflow_factory(single_instance_execution=True)
        start_step = start_step_factory(workflow=workflow, status=attached_instance.first_name)
        start_step_factory()
        process_factory(workflow=workflow, instance=attached_instance)
        wrong_instance = CompanyFactory()
        process_factory(workflow=workflow, finished=None, instance=wrong_instance)
        assert Workflow.get_start_steps_for_instance(attached_instance) == [start_step]

    def test_get_start_steps_for_instance_wrong_status_field(self, workflow_factory, start_step_factory):
        person = PersonFactory()
        workflow = workflow_factory(status_field="Test")
        start_step_factory(workflow=workflow, status=person.first_name)
        assert not Workflow.get_start_steps_for_instance(person)

    def test_get_start_steps_for_instance_wrong_status(self, workflow_factory, start_step_factory):
        person = PersonFactory()
        workflow = workflow_factory()
        start_step_factory(workflow=workflow, status="Test")
        assert not Workflow.get_start_steps_for_instance(person)

    def test_get_next_user_step_transitions_for_instance_superuser(
        self, workflow_factory, transition_factory, process_step_factory, user_step_factory
    ):
        workflow = workflow_factory()
        step = user_step_factory(workflow=workflow, permission=Permission.objects.last())
        transition = transition_factory(from_step=step, to_step__workflow=workflow)
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE)
        attached_instance = process_step.process.instance
        superuser = get_or_create_superuser()
        start_transitions = Workflow.get_next_user_step_transitions_for_instance(attached_instance, superuser)
        assert start_transitions[0][0] == [transition]
        assert str(start_transitions[0][1].pk) == process_step.pk

    def test_get_next_user_step_transitions_for_instance_wrong_instance(
        self, workflow_factory, transition_factory, process_step_factory, user_step_factory
    ):
        workflow = workflow_factory()
        instance = PersonFactory()
        step = user_step_factory(workflow=workflow, permission=Permission.objects.last())
        transition_factory(from_step=step, to_step__workflow=workflow)
        process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE)
        superuser = get_or_create_superuser()
        assert not Workflow.get_next_user_step_transitions_for_instance(instance, superuser)

    def test_get_next_user_step_transitions_for_instance_user_is_assignee(
        self, workflow_factory, transition_factory, process_step_factory, user_step_factory
    ):
        workflow = workflow_factory()
        step = user_step_factory(workflow=workflow)
        transition = transition_factory(from_step=step, to_step__workflow=workflow)
        user = UserFactory()
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE, assignee=user)
        attached_instance = process_step.process.instance
        start_transitions = Workflow.get_next_user_step_transitions_for_instance(attached_instance, user)
        assert start_transitions[0][0] == [transition]
        assert str(start_transitions[0][1].pk) == process_step.pk

    def test_get_next_user_step_transitions_for_instance_user_in_assignee_group(
        self, workflow_factory, transition_factory, process_step_factory, user_step_factory
    ):
        workflow = workflow_factory()
        step = user_step_factory(workflow=workflow, assignee_method=None)
        transition = transition_factory(from_step=step, to_step__workflow=workflow)
        group = GroupFactory()
        user = UserFactory(groups=[group])
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE, group=group)
        attached_instance = process_step.process.instance
        start_transitions = Workflow.get_next_user_step_transitions_for_instance(attached_instance, user)
        assert start_transitions[0][0] == [transition]
        assert str(start_transitions[0][1].pk) == process_step.pk

    def test_get_next_user_step_transitions_for_instance_user_in_assignee_group_but_with_method(
        self, workflow_factory, transition_factory, process_step_factory, user_step_factory
    ):
        workflow = workflow_factory()
        step = user_step_factory(workflow=workflow, permission=Permission.objects.last())
        transition_factory(from_step=step, to_step__workflow=workflow)
        group = GroupFactory()
        user = UserFactory(groups=[group])
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE, group=group)
        attached_instance = process_step.process.instance
        assert not Workflow.get_next_user_step_transitions_for_instance(attached_instance, user)

    def test_get_next_user_step_transitions_for_instance_user_has_permission(
        self, workflow_factory, transition_factory, process_step_factory, user_step_factory
    ):
        workflow = workflow_factory()
        step = user_step_factory(workflow=workflow, permission=Permission.objects.last())
        transition = transition_factory(from_step=step, to_step__workflow=workflow)
        user = UserFactory()
        user.user_permissions.add(step.permission)
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE, assignee=user)
        attached_instance = process_step.process.instance
        start_transitions = Workflow.get_next_user_step_transitions_for_instance(attached_instance, user)
        assert start_transitions[0][0] == [transition]
        assert str(start_transitions[0][1].pk) == process_step.pk

    def test_get_next_user_step_transitions_for_instance_user_has_group_permission(
        self, workflow_factory, transition_factory, process_step_factory, user_step_factory
    ):
        workflow = workflow_factory()
        step = user_step_factory(workflow=workflow, permission=Permission.objects.last())
        transition = transition_factory(from_step=step, to_step__workflow=workflow)
        group = GroupFactory(permissions=[step.permission])
        user = UserFactory(groups=[group])
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE, assignee=user)
        attached_instance = process_step.process.instance
        start_transitions = Workflow.get_next_user_step_transitions_for_instance(attached_instance, user)
        assert start_transitions[0][0] == [transition]
        assert str(start_transitions[0][1].pk) == process_step.pk

    def test_get_next_user_step_transitions_for_instance_user_has_no_permission(
        self, workflow_factory, transition_factory, process_step_factory, user_step_factory
    ):
        workflow = workflow_factory()
        step = user_step_factory(workflow=workflow, permission=Permission.objects.last())
        transition_factory(from_step=step, to_step__workflow=workflow)
        user = UserFactory()
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE, assignee=user)
        attached_instance = process_step.process.instance
        assert not Workflow.get_next_user_step_transitions_for_instance(attached_instance, user)

    def test_get_next_user_step_transitions_for_instance_user_no_permission_selected(
        self, workflow_factory, transition_factory, process_step_factory, user_step_factory
    ):
        workflow = workflow_factory()
        step = user_step_factory(workflow=workflow)
        transition = transition_factory(from_step=step, to_step__workflow=workflow)
        user = UserFactory()
        process_step = process_step_factory(step=step, state=ProcessStep.StepState.ACTIVE, assignee=user)
        attached_instance = process_step.process.instance
        start_transitions = Workflow.get_next_user_step_transitions_for_instance(attached_instance, user)
        assert start_transitions[0][0] == [transition]
        assert str(start_transitions[0][1].pk) == process_step.pk

    def test_get_next_user_step_transitions_for_instance_condition_satisfied(
        self, workflow_factory, transition_factory, process_step_factory, user_step_factory, condition_factory
    ):
        workflow = workflow_factory()
        person = PersonFactory()
        step = user_step_factory(workflow=workflow)
        transition = transition_factory(from_step=step, to_step__workflow=workflow)
        condition_factory(
            transition=transition,
            attribute_name="first_name",
            expected_value=person.first_name,
            operator=Condition.Operator.EQ,
            negate_operator=False,
        )
        superuser = get_or_create_superuser()
        process_step = process_step_factory(
            step=step, state=ProcessStep.StepState.ACTIVE, assignee=superuser, process__instance=person
        )
        start_transitions = Workflow.get_next_user_step_transitions_for_instance(person, superuser)
        assert start_transitions[0][0] == [transition]
        assert str(start_transitions[0][1].pk) == process_step.pk

    def test_get_next_user_step_transitions_for_instance_condition_not_satisfied(
        self, workflow_factory, transition_factory, process_step_factory, user_step_factory, condition_factory
    ):
        workflow = workflow_factory()
        person = PersonFactory()
        step = user_step_factory(workflow=workflow)
        transition = transition_factory(from_step=step, to_step__workflow=workflow)
        condition_factory(
            transition=transition,
            attribute_name="first_name",
            expected_value=person.first_name,
            operator=Condition.Operator.EQ,
            negate_operator=True,
        )
        superuser = get_or_create_superuser()
        process_step_factory(
            step=step, state=ProcessStep.StepState.ACTIVE, assignee=superuser, process__instance=person
        )
        assert not Workflow.get_next_user_step_transitions_for_instance(person, superuser)

    @patch("wbcore.contrib.workflow.models.workflow.Workflow.generate_workflow_png")
    def test_re_render_workflow_graph_workflow(self, mock_generate, workflow_factory):
        workflow = workflow_factory()
        old_graph = workflow.graph
        new_graph = factory.django.ImageField()._make_data({"width": 1024, "height": 768})
        mock_generate.return_value = new_graph
        re_render_workflow_graph(sender=Workflow, instance=workflow, raw=False)
        assert mock_generate.call_count == 1
        assert workflow.graph != old_graph

    @patch("wbcore.contrib.workflow.models.workflow.Workflow.generate_workflow_png")
    def test_re_render_workflow_graph_step(self, mock_generate, random_child_step_factory):
        step = random_child_step_factory()
        old_graph = step.workflow.graph
        new_graph = factory.django.ImageField()._make_data({"width": 1024, "height": 768})
        mock_generate.return_value = new_graph
        re_render_workflow_graph(sender=step.__class__, instance=step, raw=False)
        # Not really sure why its called twice but as long as we have no infinite loop I'm fine
        assert mock_generate.call_count == 2
        assert step.workflow.graph != old_graph

    @patch("wbcore.contrib.workflow.models.workflow.Workflow.generate_workflow_png")
    def test_re_render_workflow_graph_transition(self, mock_generate, transition_factory):
        transition = transition_factory()
        old_graph = transition.to_step.workflow.graph
        new_graph = factory.django.ImageField()._make_data({"width": 1024, "height": 768})
        mock_generate.return_value = new_graph
        re_render_workflow_graph(sender=Transition, instance=transition, raw=False)
        # Not really sure why its called twice but as long as we have no infinite loop I'm fine
        assert mock_generate.call_count == 2
        assert transition.to_step.workflow.graph != old_graph

    @patch("wbcore.contrib.workflow.models.workflow.Workflow.generate_workflow_png")
    def test_re_render_workflow_graph_condition(self, mock_generate, condition_factory):
        condition = condition_factory()
        old_graph = condition.transition.to_step.workflow.graph
        new_graph = factory.django.ImageField()._make_data({"width": 1024, "height": 768})
        mock_generate.return_value = new_graph
        re_render_workflow_graph(sender=Condition, instance=condition, raw=False)
        # Not really sure why its called twice but as long as we have no infinite loop I'm fine
        assert mock_generate.call_count == 2
        assert condition.transition.to_step.workflow.graph != old_graph

    @patch("wbcore.contrib.workflow.models.workflow.Workflow.generate_workflow_png")
    def test_re_render_workflow_graph_sender_renamed(self, mock_generate, workflow_factory):
        workflow = workflow_factory()
        old_graph = workflow.graph
        new_graph = factory.django.ImageField()._make_data({"width": 1024, "height": 768})
        mock_generate.return_value = new_graph
        re_render_workflow_graph(sender=ProcessStep, instance=workflow, raw=False)
        assert mock_generate.call_count == 0
        assert workflow.graph == old_graph



import importlib.abc
import importlib.machinery
import sys
from types import ModuleType


class StringSourceLoader(importlib.abc.SourceLoader):
    MOD_NAME = "__string_mod"

    def __init__(self, data):
        self.data = data.encode()

    def get_data(self, path):
        return self.data

    def get_filename(self, fullname):
        return "<string>"

    def create_module(self, spec):
        return ModuleType(spec.name)

    def exec_module(self, module):
        code = self.source_to_code(self.data, self.get_filename(module.__name__))
        exec(code, module.__dict__)  # noqa: S102

    def get_spec(self):
        return importlib.machinery.ModuleSpec(self.MOD_NAME, self, origin=self.get_filename(self.MOD_NAME))

    def register_module(self):
        sys.modules[self.MOD_NAME] = importlib._bootstrap._load(self.get_spec())

    def import_module(self):
        return importlib.import_module(self.MOD_NAME)

    def load_module(self):
        self.register_module()
        return self.import_module()

    def cleanup(self):
        del sys.modules[self.MOD_NAME]
        if self.MOD_NAME in globals():
            del globals()[self.MOD_NAME]



from wbcore.serializers.fields.text import CharField


# TODO: Color for each state
class FSMStatusField(CharField):
    field_type = "status"

    def __init__(self, *args, **kwargs):
        self.choices = kwargs.pop("choices")
        read_only = kwargs.pop("read_only", True)
        super().__init__(*args, read_only=read_only, **kwargs)

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        representation["choices"] = list()

        for choice in self.choices:
            representation["choices"].append({"value": choice[0], "label": choice[1]})

        return key, representation



from django.utils.translation import gettext as _

from wbcore.contrib.icons import WBIcon
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class TelephoneContactButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        return {
            bt.WidgetButton(
                key="new_call", label=_("Create New Call Activity"), new_mode=True, icon=WBIcon.PHONE_ADD.icon
            ),
        }

    def get_custom_instance_buttons(self):
        return self.get_custom_list_instance_buttons()



from django.utils.translation import gettext as _

from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

USER_MENUITEM = MenuItem(
    label=_("Users"),
    endpoint="wbcore:authentication:user-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["authentication.view_user"]
    ),
    add=MenuItem(
        label=_("Create New User"),
        endpoint="wbcore:authentication:user-list",
        permission=ItemPermission(permissions=["authentication.add_user"]),
    ),
)



# Generated by Django 4.1.8 on 2023-04-21 08:27

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("directory", "0003_remove_entry_last_event"),
    ]

    operations = [
        migrations.AddField(
            model_name="entry",
            name="is_draft_entry",
            field=models.BooleanField(
                default=False,
                help_text="Draft entries are entries that have been created automatically and still need verification by a human user.",
                verbose_name="Draft Entry",
            ),
        ),
    ]



import enum
from collections.abc import Iterable
from contextlib import suppress
from decimal import Decimal
from typing import Any, Dict, List, Optional, Type

import numpy as np
from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models import Model
from tqdm import tqdm

from .exceptions import DeserializationError, ImportError, SkipImportError
from .models import ImportedObjectProviderRelationship, ImportSource
from .utils import nest_row


class ImportState(enum.Enum):
    CREATED = 0
    MODIFIED = 1
    UNMODIFIED = 2


class ImportExportHandler:
    MAX_ALLOWED_LOG_SIZE: int = 2e6
    MODEL_APP_LABEL: str = None

    def __init__(self, import_source: ImportSource, **kwargs):
        self.import_source: ImportSource = import_source
        self.model: Type[Model] = apps.get_model(self.MODEL_APP_LABEL)
        self.processed_ids: list[int] = []

    def _inject_internal_id_from_data(self, data: dict[str, Any]) -> tuple[int, ContentType] | None:
        if provider_id := data.pop("provider_id", None):
            if content_type_id := data.pop("content_type", None):
                content_type = ContentType.objects.get(id=content_type_id)
            else:
                content_type = ContentType.objects.get_for_model(self.model)

            if internal_object := self.import_source.source.data_backend.get_internal_object(
                content_type, provider_id
            ):
                data["id"] = internal_object.id
            else:
                return provider_id, content_type

    def _model_dict_diff(self, model: Any, data: Dict[str, Any]) -> Dict[str, Any]:  # noqa: C901
        """
        Given a model and its dictionary representation, compare them and find out the fields that are different
        Args:
            model: The base instance
            data: The updated instance as a dictionary representation

        Returns:
            a dictionary containing the changed field and its value
        """
        change_data = dict()
        for k, v in data.items():
            if v is not None and hasattr(model, k):
                field = getattr(model, k)
                if isinstance(field, models.Manager):
                    res_list = []
                    if not isinstance(v, Iterable):
                        v = [v]
                    for vv in v:
                        if vv is not None and vv not in field.all():
                            res_list.append(vv)
                    if res_list:
                        change_data[k] = res_list
                else:
                    with suppress(FieldDoesNotExist):
                        field_obj = model._meta.get_field(k)
                        if v in [np.nan, np.inf, -np.inf]:
                            v = None
                        if v is not None:
                            if isinstance(field_obj, models.DecimalField):
                                v = round(Decimal(v), field_obj.decimal_places)
                            if isinstance(field_obj, models.FloatField):
                                v = float(v)
                            if isinstance(field_obj, models.IntegerField):
                                v = int(v)
                            if isinstance(field_obj, models.CharField):
                                v = str(v)
                            if isinstance(field, models.Model) and isinstance(v, models.Model):
                                if v.pk != field.pk:
                                    change_data[k] = v
                                # v = field.__class__.objects.get(id=v.pk) # not sure why this was there. We comment it out and monitor
                    if k not in change_data and field != v:
                        change_data[k] = v
        return change_data

    def _deserialize(self, data: Dict[str, Any]):
        """
        Convert the data from the parser to a valid model representation
        Args:
            import_source: The import source
            data: the serialized data

        Returns:
            The deserialized data
        """
        pass

    def _save_object(self, _object, **kwargs):
        save_kwargs = {}
        from wbcore.models import DynamicModel  # TODO: circular dependency

        if isinstance(_object, DynamicModel):
            save_kwargs = {
                "set_dynamic_field_on_save": getattr(self, "set_dynamic_field_on_save", True),
                "update_all_dynamic_fields": getattr(self, "update_all_dynamic_fields", True),
                "update_exclude_fields": set(
                    getattr(self, "update_exclude_fields", []) + kwargs.get("update_exclude_fields", [])
                ),
            }
        _object.save(**save_kwargs)
        return _object

    def _data_changed(
        self,
        _object,
        change_data: Dict[str, Any],
        initial_data: Dict[str, Any],
        include_update_fields: list[str] | None = None,
        exclude_update_fields: list[str] | None = None,
    ):
        """
        When data has changed, this function is called and handle the logic to update the instance
        Args:
            _object: The instance that needs to be updated
            change_data: The changed data
            initial_data: The initial data
            include_update_fields: If specified, only these fields will be updated
            exclude_update_fields: If specified, will exclude these fields from being updated
        """
        if not include_update_fields:
            include_update_fields = getattr(self, "include_update_fields", None)
        if not exclude_update_fields:
            exclude_update_fields = getattr(self, "exclude_update_fields", None)
        first_level_dict = dict()
        for k, v in change_data.items():
            if (include_update_fields is not None and k not in include_update_fields) or (
                exclude_update_fields is not None and k in exclude_update_fields and getattr(_object, k)
            ):
                pass
            else:
                if v is not None:
                    _v = getattr(_object, k)
                    if getattr(_v, "add", None) and v:
                        if isinstance(v, Iterable):
                            for elem in v:
                                if elem:
                                    _v.add(elem)
                        else:
                            _v.add(v)
                        self.import_source.log += f"\n{k}: Add {v} to manytomany list"
                    else:
                        setattr(_object, k, v)
                        self.import_source.log += f"\n{k}: {_v} => {v}"
                    first_level_dict[k] = v
        if first_level_dict:
            _object.import_source = self.import_source
            try:
                self._save_object(_object, update_exclude_fields=list(initial_data.keys()))
            except Exception as e:
                error_msg = f"\nError {e} while saving data {change_data} for object id {_object.pk}"
                self.import_source.log += error_msg
                if not getattr(self, "allow_update_save_failure", False):
                    raise ImportError(error_msg) from e
            return True
        return False

    def _create_instance(self, data: Dict[str, Any], **kwargs) -> models.Model:
        """
        Resource function to create the object given its deserialized data
        Args:
            data: The deserialized data

        Returns:
            The created object
        """
        self.import_source.log += f"\nCreate {self.model.__class__.__name__}."
        _object = self.model(import_source=self.import_source, **data)
        return self._save_object(_object)

    def _get_instance(self, data: Dict[str, Any], history: Optional[models.QuerySet] = None, **kwargs) -> Any | None:
        """
        Usually overriden by the resource to define how the object needs to be query given the deserialized data
        Args:
            data: deserialized data
            history: In case of historical import, specify a dictionary of history datapoints

        Returns:
            Thre retreived object (if it exists), None otherwise
        """
        return None

    def _get_history(self, history: Dict[str, str]) -> models.QuerySet | None:
        return self.model.objects.none()

    def _pre_processing_object(self, data: Dict[str, str]):
        pass

    def _process_raw_data(self, data: Dict[str, Any]):
        pass

    def _post_processing_created_object(self, _object: models.Model):
        pass

    def _post_processing_updated_object(self, _object: models.Model):
        pass

    def _post_processing_history(self, history: models.QuerySet):
        pass

    def _post_processing_objects(
        self,
        created_objs: List[models.Model],
        modified_objs: List[models.Model],
        unmodified_objs: List[models.Model],
    ):
        pass

    def process_object(
        self,
        data: Dict[str, Any],
        history: Optional[models.QuerySet] = None,
        read_only=False,
        include_update_fields=None,
        raise_exception: bool = True,
        **kwargs,
    ):
        data = nest_row(data)
        inject_internal_id_res = self._inject_internal_id_from_data(data)
        if history is None:
            history = self.model.objects.none()
        self._deserialize(data)
        if not data:
            raise DeserializationError("Data dictionary is empty")
        self._pre_processing_object(data)
        import_state = ImportState.UNMODIFIED
        _object = self._get_instance(data, history=history, **kwargs)
        if _object:
            change_data = self._model_dict_diff(_object, data)
            different = len(change_data.keys()) > 0
            if different and not read_only:
                is_modified = self._data_changed(
                    _object, change_data, data, include_update_fields=include_update_fields
                )
                if is_modified:
                    import_state = ImportState.MODIFIED
                self._post_processing_updated_object(_object)
        elif not read_only:
            _object = self._create_instance(data, **kwargs)
            self._post_processing_created_object(_object)
            import_state = ImportState.CREATED
        if not _object and raise_exception:
            data_repr = " ".join([f'{k}="{v}"' for k, v in data.items()])
            raise DeserializationError(f"{self.model._meta.verbose_name} data couldn't be parsed ({data_repr})")
        if inject_internal_id_res:
            ImportedObjectProviderRelationship.objects.get_or_create(
                object_id=_object.pk,
                content_type=inject_internal_id_res[1],
                provider=self.import_source.source.data_backend.provider,
                defaults={"provider_identifier": inject_internal_id_res[0]},
            )

        return _object, import_state

    def process(self, import_data, debug: bool = False, with_post_processing: bool = True, **kwargs):
        data = import_data["data"]
        self._process_raw_data(import_data)

        if history_data := import_data.get("history", None):
            history = self._get_history(history_data)
        else:
            history = self.model.objects.none()

        created_objs = []
        modified_objs = []
        unmodified_objs = []
        gen = data if not debug else tqdm(data, total=len(data))
        for _data in gen:
            try:
                _object, import_state = self.process_object(_data, history=history)
                if history.exists() and _object:
                    history = history.exclude(id=_object.pk)

                if import_state == ImportState.CREATED:
                    created_objs.append(_object)
                elif import_state == ImportState.MODIFIED:
                    modified_objs.append(_object)
                else:
                    unmodified_objs.append(_object)
                if (
                    len(self.import_source.log) > self.MAX_ALLOWED_LOG_SIZE
                ):  # In case we are exceeding the max log size, we reset the log to avoid issue when saving it
                    self.import_source.log = ""
                if object_id := getattr(_object, "id", None):
                    self.processed_ids.append(object_id)
            except SkipImportError as e:
                self.import_source.log += f"skipping Row {self.import_source.progress_index}: {str(e)}\n"
            except DeserializationError as e:
                self.import_source.errors_log += f"Warning Row {self.import_source.progress_index}: {str(e)}\n"
            self.import_source.progress_index += 1
        if with_post_processing:
            if history.exists():
                self._post_processing_history(history)
            self._post_processing_objects(created_objs, modified_objs, unmodified_objs)

        self.import_source.save()



import factory
from django.db.models import signals

from wbcore.contrib.icons import WBIcon
from wbcore.contrib.workflow.models import Transition


@factory.django.mute_signals(signals.post_save)
class TransitionFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Transition
        exclude = ("_workflow",)

    _workflow = factory.SubFactory("wbcore.contrib.workflow.factories.WorkflowFactory")

    from_step = factory.SubFactory(
        "wbcore.contrib.workflow.factories.StartStepFactory",
        workflow=factory.SelfAttribute(".._workflow"),
    )
    to_step = factory.SubFactory(
        "wbcore.contrib.workflow.factories.FinishStepFactory",
        workflow=factory.SelfAttribute(".._workflow"),
    )

    icon = factory.Iterator(WBIcon.values)

    name = factory.Sequence(lambda n: f"Transition #{n}")



from django.db.models import Exists, OuterRef, Q
from django.utils.translation import gettext_lazy as _
from django_filters.widgets import CSVWidget

from wbcore import filters as wb_filters
from wbcore.contrib.agenda.models import CalendarItem
from wbcore.contrib.geography.models import Geography

from ..models import (
    Company,
    CompanyType,
    CustomerStatus,
    EmployerEmployeeRelationship,
    Entry,
    Person,
    Position,
)


class EntryFilter(wb_filters.FilterSet):
    tier = wb_filters.MultipleChoiceFilter(
        label=_("Tier"),
        help_text=_("Filter by tiering (1-5, with 1 being the best)"),
        choices=Company.Tiering.choices,
        widget=CSVWidget,
    )

    last_event_tmp = wb_filters.DateTimeRangeFilter(
        label=_("Has Event between.."),
        help_text=_("Filter by activities in a date range"),
        method="filter_last_event",
    )
    last_event_period_endswith = wb_filters.DateTimeFilter(
        label=_("Last Event Period End date"), field_name="last_event_period_endswith", lookup_expr="exact"
    )
    last_event_period_endswith__gte = wb_filters.DateTimeFilter(
        label=_("Last Event Period End date"), field_name="last_event_period_endswith", lookup_expr="gte"
    )
    last_event_period_endswith__lte = wb_filters.DateTimeFilter(
        label=_("Last Event Period End date"), field_name="last_event_period_endswith", lookup_expr="lte"
    )

    activity_heat = wb_filters.MultipleLookupFilter(
        field_class=wb_filters.RangeSelectFilter,
        lookup_expr=["gte", "lte"],
        label=_("Activity Heat"),
        help_text=_("Filter by the recent activity level"),
        precision=2,
        field_name="activity_heat",
    )

    # activity_heat__gte = wb_filters.NumberFilter(
    #     label=_("Activity Heat"),
    #     help_text=_("Filter by the recent activity level"),
    #     lookup_expr="gte",
    #     precision=2,
    #     field_name="activity_heat",
    # )
    # activity_heat__lte = wb_filters.NumberFilter(
    #     label=_("Activity Heat"),
    #     help_text=_("Filter by the recent activity level"),
    #     lookup_expr="lte",
    #     precision=2,
    #     field_name="activity_heat",
    # )

    cities = wb_filters.ModelChoiceFilter(
        method="filter_city",
        label=_("City"),
        help_text=_("Filter by related city"),
        queryset=Geography.objects.all(),
        endpoint=Geography.get_representation_endpoint(),
        value_key=Geography.get_representation_value_key(),
        label_key=Geography.get_representation_label_key(),
        filter_params={"notnull_related_name": "contact_city"},
    )

    full_addresses = wb_filters.ModelChoiceFilter(
        method="filter_full_address",
        label=_("Full Address"),
        help_text=_("Filter by an address or geography (e.g. Geneva/Switzerland/Europe)"),
        queryset=Geography.objects.all(),
        endpoint=Geography.get_representation_endpoint(),
        value_key=Geography.get_representation_value_key(),
        label_key=Geography.get_representation_label_key(),
    )

    def filter_last_event(self, queryset, label, value):
        if value:
            return queryset.annotate(
                has_events=Exists(CalendarItem.objects.filter(entities=OuterRef("pk"), period__overlap=value))
            ).filter(has_events=True)
        return queryset

    def filter_city(self, queryset, name, value):
        if value:
            return queryset.filter(addresses__geography_city=value).distinct()
        return queryset

    def filter_full_address(self, queryset, name, value):
        if value:
            city_level: int = 3
            return queryset.filter(
                **{f'addresses__geography_city{"__parent" * (city_level - value.level)}': value}
            ).distinct()
        return queryset

    class Meta:
        model = Entry
        fields = {}


class PersonFilter(EntryFilter):
    name = wb_filters.CharFilter(label=_("Name"), lookup_expr="icontains")

    position_in_company = wb_filters.ModelMultipleChoiceFilter(
        label=_("Positions"),
        queryset=Position.objects.all(),
        endpoint=Position.get_representation_endpoint(),
        value_key=Position.get_representation_value_key(),
        label_key=Position.get_representation_label_key(),
    )
    customer_status = wb_filters.ModelMultipleChoiceFilter(
        label=_("Customer Statuses"),
        queryset=CustomerStatus.objects.all(),
        endpoint=CustomerStatus.get_representation_endpoint(),
        value_key=CustomerStatus.get_representation_value_key(),
        label_key=CustomerStatus.get_representation_label_key(),
        method="filter_status",
    )
    primary_telephone = wb_filters.CharFilter(label=_("Telephone"), method="filter_primary_telephone")
    email = wb_filters.CharFilter(label=_("Email"), method="filter_email")

    only_internal_users = wb_filters.BooleanFilter(label=_("Only Internal Users"), method="filter_only_internal_users")
    with_active_user_account = wb_filters.BooleanFilter(
        label=_("With Active User Account"), method="filter_with_active_user_account"
    )

    def filter_only_internal_users(self, queryset, name, value):
        if value:
            return queryset.filter_only_internal()
        return queryset

    def filter_with_active_user_account(self, queryset, name, value):
        if value:
            return queryset.filter(user_account__isnull=False, user_account__is_active=True)
        return queryset

    def filter_address(self, queryset, name, value):
        if value:
            city_level: int = 3
            filter_key = f'addresses__geography_city{"__parent" * (city_level - value.level)}'
            return queryset.filter(
                Q(**{filter_key: value})
                | Q(
                    Exists(
                        EmployerEmployeeRelationship.objects.filter(
                            employee=OuterRef("pk"), **{"employer__" + filter_key: value}
                        )
                    )
                )
            ).distinct()
        return queryset

    def filter_status(self, queryset, name, value):
        if value:
            status_title_list = [x.title for x in value]
            return queryset.filter(customer_status__in=status_title_list)
        return queryset

    def filter_email(self, queryset, name, value):
        if value:
            return queryset.filter(emails__address__icontains=value).distinct()
        return queryset

    def filter_primary_telephone(self, queryset, name, value):
        if value:
            return queryset.filter(primary_telephone__icontains=value).distinct()
        return queryset

    def get_union_employee(self, queryset, name, value):
        if (profile := getattr(self.request.user, "profile", None)) and profile.employers.exists():
            return queryset.filter(
                Q(employers__in=[value]) | Q(employers__in=profile.employers.values("id"))
            ).distinct()
        return queryset.filter(employers__in=[value])

    class Meta:
        model = Person
        fields = {
            "employers": ["exact"],
            "relationship_managers": ["exact"],
            "specializations": ["exact"],
        }


class CompanyFilter(EntryFilter):
    type = wb_filters.ModelMultipleChoiceFilter(
        label=_("Type"),
        help_text=_("Filter by the type of company"),
        queryset=CompanyType.objects.all(),
        endpoint=CompanyType.get_representation_endpoint(),
        value_key=CompanyType.get_representation_value_key(),
        label_key=CompanyType.get_representation_label_key(),
    )

    customer_status = wb_filters.ModelMultipleChoiceFilter(
        label=_("Status"),
        help_text=_("Filter by the status of a company"),
        queryset=CustomerStatus.objects.all(),
        endpoint=CustomerStatus.get_representation_endpoint(),
        value_key=CustomerStatus.get_representation_value_key(),
        label_key=CustomerStatus.get_representation_label_key(),
    )

    employees = wb_filters.ModelChoiceFilter(
        label=_("Employees"),
        help_text=_("Filter by some person who is an employee of the company"),
        queryset=lambda request: Person.objects.exclude(
            id__in=Person.objects.filter_only_internal()
        ),  # we need to lazy load queryset so that it is not evaluated at runtime
        endpoint=Person.get_representation_endpoint(),
        value_key=Person.get_representation_value_key(),
        label_key=Person.get_representation_label_key(),
    )
    help_texts = {
        "name": "Filter by the name of the company",
        "relationship_managers": "Filter by a person who is the relationship manager of a company",
    }

    class Meta:
        model = Company
        fields = {
            "name": ["exact", "icontains"],
            "relationship_managers": ["exact"],
            "tier": ["exact", "lt", "lte", "gt", "gte"],
        }


class UserIsManagerFilter(wb_filters.FilterSet):
    computed_str = wb_filters.CharFilter(label=_("Name"), lookup_expr="icontains")
    primary_email = wb_filters.CharFilter(label=_("Email"), method="filter_primary_email")
    primary_telephone = wb_filters.CharFilter(label=_("Telephone"), method="filter_primary_telephone")

    primary_address = wb_filters.ModelChoiceFilter(
        method="filter_address",
        label=_("Address"),
        queryset=Geography.objects.all(),
        endpoint=Geography.get_representation_endpoint(),
        value_key=Geography.get_representation_value_key(),
        label_key=Geography.get_representation_label_key(),
    )

    def filter_address(self, queryset, name, value):
        if value:
            return queryset.filter(
                addresses__geography_city__in=value.get_descendants(include_self=True), addresses__primary=True
            )
        return queryset

    def filter_primary_email(self, queryset, name, value):
        if value:
            return queryset.filter(primary_email__icontains=value)
        return queryset

    def filter_primary_telephone(self, queryset, name, value):
        if value:
            return queryset.filter(primary_telephone__icontains=value)
        return queryset

    class Meta:
        model = Entry
        fields = {}



from django.utils.translation import gettext as _

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import Display, Layout, Page, Style
from wbcore.metadata.configs.display.instance_display.enums import NavigationType
from wbcore.metadata.configs.display.instance_display.operators import default, lte
from wbcore.metadata.configs.display.list_display import ListDisplay
from wbcore.metadata.configs.display.view_config import DisplayViewConfig

SEASON_FIELDS = [
    dp.Field(key="name", label=_("Season Name")),
    dp.Field(key="league", label=_("League")),
    dp.Field(key="date_range", label=_("Range")),
    dp.Field(key="duration", label=_("Duration")),
    dp.Field(key="winner", label=_("Winner")),
    dp.Field(key="top_scorer", label=_("Top Scorer")),
]


class SeasonDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return dp.ListDisplay(
            fields=(
                dp.Field(key="name", label=_("Season Name")),
                dp.Field(key="league", label=_("League")),
                dp.Field(key="winner", label=_("Winner")),
                dp.Field(key="top_scorer", label=_("Top Scorer")),
                dp.Field(key="date_range", label=_("Range")),
            )
        )

    def get_instance_display(self) -> Display:
        return Display(
            navigation_type=NavigationType.PAGE,
            pages=[
                Page(
                    title=_("Season"),
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["name", "league"],
                                ["date_range", "date_range"],
                                ["winner", "top_scorer"],
                                ["file", "."],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                        ),
                        lte(550): Layout(
                            grid_template_areas=[
                                ["name"],
                                ["league"],
                                ["date_range"],
                                ["winner"],
                                ["top_scorer"],
                                ["file"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                        ),
                    },
                ),
            ],
        )



from rest_framework.permissions import IsAuthenticated


class IsClientManagerRelationshipAdmin(IsAuthenticated):
    def has_permission(self, request, view):
        return request.user.has_perm("directory.administrate_clientmanagerrelationship")



from django.contrib import admin

from ..models import (
    AddressContact,
    BankingContact,
    EmailContact,
    SocialMediaContact,
    TelephoneContact,
    WebsiteContact,
)


@admin.register(TelephoneContact)
class TelephoneContactModelAdmin(admin.ModelAdmin):
    autocomplete_fields = ("entry",)
    search_fields = ("entry__computed_str", "number")


@admin.register(WebsiteContact)
class WebsiteContactModelAdmin(admin.ModelAdmin):
    autocomplete_fields = ("entry",)
    search_fields = ("entry__computed_str", "url")


@admin.register(EmailContact)
class EmailContactModelAdmin(admin.ModelAdmin):
    autocomplete_fields = ("entry",)
    search_fields = ("entry__computed_str", "address")


@admin.register(AddressContact)
class AddressContactModelAdmin(admin.ModelAdmin):
    autocomplete_fields = ("entry",)
    search_fields = (
        "entry__computed_str",
        "street",
        "street_additional",
        "zip",
        "geography_city__name",
        "geography_city__parent__name",
        "geography_city__parent__parent__name",
    )
    raw_id_fields = ["geography_city"]


@admin.register(SocialMediaContact)
class SocialMediaContact(admin.ModelAdmin):
    autocomplete_fields = ("entry",)
    search_fields = ("entry__computed_str", "platform")


@admin.register(BankingContact)
class BankingContactModelAdmin(admin.ModelAdmin):
    autocomplete_fields = ("entry",)
    search_fields = ("entry__computed_str", "institute", "institute_additional", "iban", "swift_bic")



from contextlib import suppress
from datetime import datetime
from typing import TYPE_CHECKING, Iterable, Iterator

from django.contrib.contenttypes.models import ContentType
from django.db import ProgrammingError
from django.db.models import Model, Q, QuerySet
from django.utils import timezone
from guardian.shortcuts import assign_perm, get_anonymous_user
from psycopg.errors import InvalidCursorName
from wbcore.contrib.authentication.models import User
from wbcore.contrib.guardian.models.models import UserObjectPermission
from wbcore.permissions.shortcuts import get_internal_users

if TYPE_CHECKING:
    from wbcore.contrib.guardian.models.mixins import PermissionObjectModelMixin


def assign_permissions(permissions_map: Iterable[tuple[str, Model, User, bool]]):
    """
    Assigns object-level permissions to users based on the provided permissions map.

    This method iterates through the permissions_map, assigns the specified permission
    to the user for the model instance, sets the 'editable' flag for the permission,
    and saves the object permission.
    """
    for permission, instance, user, editable in permissions_map:
        object_permissions = assign_perm(permission, user, instance)

        if object_permissions is None:
            continue

        if isinstance(object_permissions, Model):
            object_permissions = [object_permissions]

        for object_permission in object_permissions:
            object_permission.editable = editable  # type: ignore -- We have our custom Permission class here
            object_permission.save()


def get_public_user_or_group(only_internal: bool = False) -> QuerySet[User]:
    """
    Retrieves a queryset of active users, optionally filtering to internal users only.

    This method fetches users based on the `only_internal` flag.
    If set, it filters for internal users, otherwise returns all users.
    The queryset includes active users or the anonymous user and excludes superusers
    from the main set, while later ensuring their inclusion through a union operation.
    """
    users = User.objects.filter(is_active=True)
    if only_internal:
        users = users.filter(
            Q(is_superuser=True) | Q(id=get_anonymous_user().pk) | Q(id__in=get_internal_users().values("id"))
        )
    return users


def get_permission_matrix(
    queryset: QuerySet,
    created: datetime | None = None,
    instance: Model | None = None,
    user: User | None = None,
) -> Iterator[tuple[str, Model, User, bool]]:
    """
    Retrieves the permission matrix for all (user, object) pairs

    If an instance is provided, the queryset is filtered to that specific object. The method determines the
    appropriate set of users based on the permission type of the object (private or public).

    For each user, the function yields a tuple containing the permission string, the object instance,
    the user, and whether the permission is editable.
    """

    if instance:
        queryset = queryset.filter(id=instance.id)  # type: ignore

    for _instance in queryset.all():
        if _instance.permission_type is _instance.PermissionType.PUBLIC:
            users = get_public_user_or_group()
        else:
            users = get_public_user_or_group(only_internal=True)
        if user:
            users = users.filter(id=user.id)
        for user in users:
            for permission, editable in _instance.get_permissions_for_user(user, created=created).items():
                yield permission, _instance, user, editable


def prune_permissions(instance: "PermissionObjectModelMixin", force: bool | None = False):
    queryset = UserObjectPermission.objects.filter(
        content_type=ContentType.objects.get_for_model(instance), object_pk=instance.id
    )
    if not force:
        queryset = queryset.exclude(editable=True)
    for permission in queryset:
        permission.delete()


def reload_permissions(
    queryset: QuerySet,
    created: datetime | None = None,
    user: User | None = None,
    instance: "PermissionObjectModelMixin | None" = None,
    prune_existing: bool = True,
    force_pruning: bool = False,
):
    """
    Assigns permissions based on a given queryset, with options to prune existing permissions and
    specify the creation timestamp. If no creation time is provided, the current time is used.

    The function first checks if existing permissions should be pruned, which happens if both
    `prune_existing` and `instance` are provided. It then retrieves the permission matrix
    using `get_permission_matrix()` and assigns the appropriate permissions.

    Error handling is in place to suppress database-related errors, such as `ProgrammingError`
    and `InvalidCursorName`, which can occur due to unmanaged tables.
    """
    if not created:
        created = timezone.now()
    with suppress(ProgrammingError, InvalidCursorName):  # We check this to catch error trigger by unmanaged table
        if prune_existing and instance:
            prune_permissions(instance, force=force_pruning)
        permission_matrix = get_permission_matrix(queryset, created=created, instance=instance, user=user)
        assign_permissions(permission_matrix)



from django.contrib import admin

from wbcore.contrib.workflow.models import Workflow


@admin.register(Workflow)
class WorkflowAdmin(admin.ModelAdmin):
    search_fields = ("name",)
    list_display = ("id", "name", "single_instance_execution", "model", "status_field", "preserve_instance")



from enum import Enum
from typing import Union


class RequestType(Enum):
    GET = "get"
    POST = "post"
    DELETE = "delete"
    PATCH = "patch"
    PUT = "put"
    OPTIONS = "options"
    HEAD = "head"


class WidgetType(Enum):
    LIST = "list"
    INSTANCE = "instance"
    CHART = "chart"
    SELECT = "select"
    HTML = "html"


class Operator(Enum):
    EQUAL = "=="
    UNEQUAL = "!="
    GREATER = ">"
    GREATER_EQUAL = ">="
    LESS = "<"
    LESS_EQUAL = "<="
    EXISTS = "∃"


class Unit(Enum):
    FRACTION = "fr"
    REM = "rem"
    PIXEL = "px"

    def __call__(self, _value):
        return (float(_value), self.value)

    def unit(self, _value: Union[float, str, int]):
        if not isinstance(_value, (float, str, int)):
            raise AssertionError("_value needs to be one of str, float or int")

        return f"{float(_value)}{self.value}"


class AuthType(Enum):
    NONE = "NONE"
    JWT = "JWT"



from django.dispatch import receiver
from dynamic_preferences.registries import global_preferences_registry

from wbcore.test.signals import custom_update_kwargs, get_custom_factory

from ..factories import (
    BankFactory,
    ParentRelationshipTypeFactory,
    RandomClientFactory,
    UserIsManagerEntryFactory,
)
from ..viewsets import (
    BankModelViewSet,
    EmployeeEmployerModelViewSet,
    EmployerEmployeeModelViewSet,
    PersonInChargeRepresentationViewSet,
    RelationshipModelEntryViewSet,
    RelationshipTypeModelViewSet,
    SystemEmployeeModelViewSet,
    UserIsManagerViewSet,
)

# =================================================================================================================
#                                              CUSTOM FACTORY
# =================================================================================================================


@receiver(get_custom_factory, sender=PersonInChargeRepresentationViewSet)
def receive_factory_person_in_charge(sender, *args, **kwargs):
    return RandomClientFactory


@receiver(get_custom_factory, sender=RelationshipTypeModelViewSet)
def receive_factory_relationship_type(sender, *args, **kwargs):
    return ParentRelationshipTypeFactory


@receiver(get_custom_factory, sender=BankModelViewSet)
def receive_factory_bank(sender, *args, **kwargs):
    return BankFactory


@receiver(get_custom_factory, sender=UserIsManagerViewSet)
def receive_factory_user_is_manager(sender, *args, **kwargs):
    return UserIsManagerEntryFactory


# =================================================================================================================
#                                              UPDATE KWARGS
# =================================================================================================================


@receiver(custom_update_kwargs, sender=RelationshipModelEntryViewSet)
def receive_kwargs_relationship(sender, *args, **kwargs):
    if kwargs.get("from_entry_id"):
        return {"entry_id": kwargs.get("from_entry_id")}
    return {}


@receiver(custom_update_kwargs, sender=EmployeeEmployerModelViewSet)
def receive_kwargs_employee_company(sender, *args, **kwargs):
    if obj := kwargs.get("obj_factory"):
        return {"company_id": obj.employer.id}
    return {}


@receiver(custom_update_kwargs, sender=EmployerEmployeeModelViewSet)
def receive_kwargs_employer_person(sender, *args, **kwargs):
    if obj := kwargs.get("obj_factory"):
        return {"employee_id": obj.employee.id}
    return {}


@receiver(custom_update_kwargs, sender=SystemEmployeeModelViewSet)
def receive_kwargs_system_employee(sender, *args, **kwargs):
    if obj := kwargs.get("obj_factory"):
        company_id = obj.employer.id
        global_preferences_registry.manager()["directory__main_company"] = company_id
        return {"employer_id": company_id}
    return {}


@receiver(custom_update_kwargs, sender=UserIsManagerViewSet)
def receive_kwargs_user_is_manager(sender, *args, **kwargs):
    if (obj := kwargs.get("obj_factory")) and (user := kwargs.get("user")):
        if cmr := obj.client_of.first():
            cmr.relationship_manager = user.profile
            cmr.save()
    return {}



from django.db.models import Model
from django.http import HttpResponse
from django.test import Client
from django.urls import reverse
from wbcore.contrib.authentication.models import User
from wbcore.serializers import Serializer
from wbcore.test.utils import get_data_from_factory
from wbcore.viewsets import ModelViewSet


def get_create_view(
    client: Client,
    instance: Model,
    superuser: User,
    url: str,
    viewset: ModelViewSet,
) -> HttpResponse:
    """
    Create a new instance through a view using the Django REST Framework client.

    Parameters:
        - client (django.test.Client): The Django test client.
        - instance (django.db.models.Model): The model instance to be created.
        - superuser (django.contrib.auth.models.User): The superuser for authentication.
        - url (str): The URL endpoint for the view.
        - viewset (wbcore.viewsets.ModelViewSet): The viewset for the model.

    Returns:
        django.http.HttpResponse: The HTTP response from the view.
    """
    instance_data = get_data_from_factory(instance, viewset, delete=True, superuser=superuser)
    return client.post(url, instance_data)


def get_detail_view(client: Client, pk: int, url: str) -> HttpResponse:
    """
    Retrieve the details of a model instance through a detail view using the Django REST Framework client.

    Parameters:
        - client (django.test.Client): The Django test client.
        - pk (int): The primary key of the model instance to retrieve.
        - url (str): The base URL endpoint for the detail view.

    Returns:
        django.http.HttpResponse: The HTTP response from the detail view.
    """
    detail_url = reverse(url, args=[pk])
    return client.get(detail_url)


def get_update_view(client: Client, instance: Model, serializer: Serializer, url: str) -> HttpResponse:
    """
    Update a model instance through an update view using the Django REST Framework client.

    Parameters:
        - client (django.test.Client): The Django test client.
        - instance (django.db.models.Model): The model instance to be updated.
        - serializer (Serializer): The serializer used to serialize the instance.
        - url (str): The base URL endpoint for the update view.

    Returns:
        django.http.HttpResponse: The HTTP response from the update view.
    """
    update_url = reverse(url, args=[instance.pk])
    serialized_data = serializer(instance).data
    return client.put(update_url, data=serialized_data, content_type="application/json")


def get_partial_view(client: Client, instance_id: int, data: dict, url: str) -> HttpResponse:
    """
    Facilitates partial updates to a specific instance of a model using the Django REST Framework client.

    Parameters:
        - client (django.test.Client): The Django test client.
        - instance_id (int): The unique identifier (primary key) of the instance to be partially updated.
        - data (dict): A dictionary containing the fields and their respective new values that need to be updated in the instance.
        - url (str): The base URL endpoint for the patch view.
    Returns:
        django.http.HttpResponse: The HTTP response from the patch view.
    """
    update_url = reverse(url, args=[instance_id])
    return client.patch(update_url, data=data, content_type="application/json")


def get_delete_view(client: Client, url: str, pk: int) -> HttpResponse:
    """
    Delete a model instance through a delete view using the Django REST Framework client.

    Parameters:
        - client (django.test.Client): The Django test client.
        - url (str): The base URL endpoint for the delete view.
        - pk (int): The primary key of the model instance to delete.

    Returns:
        django.http.HttpResponse: The HTTP response from the delete view.
    """
    delete_url = reverse(url, args=[pk])
    return client.delete(delete_url)


def find_instances_in_response(instances: list[Model], response: HttpResponse) -> tuple:
    """
    Find instances in the data contained in an HTTP response.

    This method takes a list of instances and an HTTP response object, and it checks if each instance
    is present in the data of the response based on their unique identifier (e.g., 'id').

    Parameters:
        - instances (list): A list of instances (models) to search for in the response data.
        - response (HttpResponse): An HTTP response object containing data to search within.

    Returns:
        tuple: A tuple of Boolean values indicating whether each instance was found in the response.
            Each element in the tuple corresponds to an instance in the same order as in the 'instances' list.
            If an instance is found, the corresponding value is True; otherwise, it is False.
    """
    found_instances = [False] * len(instances)

    for item in response.data["results"]:
        for index, instance in enumerate(instances):
            if item.get("id") == instance.id:
                found_instances[index] = True

    return tuple(found_instances)



from wbcore.serializers.fields.text import CharField
from wbcore.serializers.fields.types import WBCoreType

from .icons import WBIcon


class IconSelectField(CharField):
    field_type = WBCoreType.ICON.value

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def to_representation(self, value):
        # Try to get its "icon" representation. If the field is not supported, we simply returns the value (This allow custom and hardcoded icon)
        try:
            return WBIcon[value].icon
        except KeyError:
            return value



from django.utils.translation import gettext_lazy as _

from wbcore import filters as wb_filters
from wbcore.contrib.authentication.models import Group, Permission, User
from wbcore.contrib.workflow.models import Process, ProcessStep, Step, Workflow


class ProcessWorkflowFilter(wb_filters.FilterSet):
    started__lte = wb_filters.DateTimeFilter(
        label=_("Started Before/On"),
        lookup_expr="lte",
        field_name="started",
    )
    started__gte = wb_filters.DateTimeFilter(
        label=_("Started After/On"),
        lookup_expr="gte",
        field_name="started",
    )
    finished__lte = wb_filters.DateTimeFilter(
        label=_("Finished Before/On"),
        lookup_expr="lte",
        field_name="finished",
    )
    finished__gte = wb_filters.DateTimeFilter(
        label=_("Finished After/On"),
        lookup_expr="gte",
        field_name="finished",
    )
    id = wb_filters.CharFilter(label=_("UUID"))
    state = wb_filters.MultipleChoiceFilter(choices=Process.ProcessState.choices, label=_("States"))

    class Meta:
        model = Process
        fields = {}


class ProcessFilter(ProcessWorkflowFilter):
    workflow = wb_filters.ModelMultipleChoiceFilter(
        label=_("Workflows"),
        queryset=Workflow.objects.all(),
        endpoint=Workflow.get_representation_endpoint(),
        value_key=Workflow.get_representation_value_key(),
        label_key=Workflow.get_representation_label_key(),
    )

    class Meta(ProcessWorkflowFilter.Meta):
        model = Process
        fields = ProcessWorkflowFilter.Meta.fields


class ProcessStepBaseFilter(ProcessWorkflowFilter):
    assignee = wb_filters.ModelMultipleChoiceFilter(
        label=_("Assignees"),
        queryset=User.objects.all(),
        endpoint=User.get_representation_endpoint(),
        value_key=User.get_representation_value_key(),
        label_key=User.get_representation_label_key(),
    )
    group = wb_filters.ModelMultipleChoiceFilter(
        label=_("Groups"),
        queryset=Group.objects.all(),
        endpoint=Group.get_representation_endpoint(),
        value_key=Group.get_representation_value_key(),
        label_key=Group.get_representation_label_key(),
    )
    permission = wb_filters.ModelMultipleChoiceFilter(
        label=_("Permissions"),
        queryset=Permission.objects.all(),
        endpoint=Permission.get_representation_endpoint(),
        value_key=Permission.get_representation_value_key(),
        label_key=Permission.get_representation_label_key(),
    )

    class Meta:
        model = ProcessStep
        fields = {
            "status": ["exact", "icontains"],
            "error_message": ["exact", "icontains"],
        }


class ProcessStepProcessFilter(ProcessStepBaseFilter):
    step = wb_filters.ModelMultipleChoiceFilter(
        label=_("Steps"),
        queryset=Step.objects.all(),
        endpoint=Step.get_representation_endpoint(),
        value_key=Step.get_representation_value_key(),
        label_key=Step.get_representation_label_key(),
    )

    class Meta(ProcessStepBaseFilter.Meta):
        model = ProcessStep
        fields = ProcessStepBaseFilter.Meta.fields


class ProcessStepStepFilter(ProcessStepBaseFilter):
    process = wb_filters.ModelMultipleChoiceFilter(
        label=_("Processes"),
        queryset=Process.objects.all(),
        endpoint=Process.get_representation_endpoint(),
        value_key=Process.get_representation_value_key(),
        label_key=Process.get_representation_label_key(),
    )

    class Meta(ProcessStepBaseFilter.Meta):
        model = ProcessStep
        fields = ProcessStepBaseFilter.Meta.fields


class ProcessStepFilter(ProcessStepProcessFilter, ProcessStepStepFilter):
    class Meta(ProcessStepBaseFilter.Meta):
        model = ProcessStep
        fields = ProcessStepBaseFilter.Meta.fields


class AssignedProcessStepFilter(wb_filters.FilterSet):
    step = wb_filters.ModelMultipleChoiceFilter(
        label=_("Steps"),
        queryset=Step.objects.all(),
        endpoint=Step.get_representation_endpoint(),
        value_key=Step.get_representation_value_key(),
        label_key=Step.get_representation_label_key(),
    )
    started__lte = wb_filters.DateTimeFilter(
        label=_("Started Before/On"),
        lookup_expr="lte",
        field_name="started",
    )
    started__gte = wb_filters.DateTimeFilter(
        label=_("Started After/On"),
        lookup_expr="gte",
        field_name="started",
    )
    finished__lte = wb_filters.DateTimeFilter(
        label=_("Finished Before/On"),
        lookup_expr="lte",
        field_name="finished",
    )
    finished__gte = wb_filters.DateTimeFilter(
        label=_("Finished After/On"),
        lookup_expr="gte",
        field_name="finished",
    )
    group = wb_filters.ModelMultipleChoiceFilter(
        label=_("Groups"),
        queryset=Group.objects.all(),
        endpoint=Group.get_representation_endpoint(),
        value_key=Group.get_representation_value_key(),
        label_key=Group.get_representation_label_key(),
    )
    permission = wb_filters.ModelMultipleChoiceFilter(
        label=_("Permissions"),
        queryset=Permission.objects.all(),
        endpoint=Permission.get_representation_endpoint(),
        value_key=Permission.get_representation_value_key(),
        label_key=Permission.get_representation_label_key(),
    )
    workflow_name = wb_filters.CharFilter(label=_("Workflow"), lookup_expr="icontains")
    attached_model = wb_filters.CharFilter(label=_("Attached Model"), lookup_expr="icontains")

    class Meta:
        model = ProcessStep
        fields = {
            "status": ["exact", "icontains"],
        }



from .team import TeamFilter, TeamStadiumFilter
from .person import PlayerTeamFilter, PlayerFilter, SportPersonFilter, SportPersonRepresentationFilter
from .role import RoleFilter
from .stadium import StadiumFilter
from .league import LeagueFilter, LeagueSportFilter
from .sport import SportFilter
from .match import MatchFilter, MatchLeagueFilter, MatchStadiumFilter
from .event import EventFilter, EventTypeFilter, EventTypeSportFilter, EventMatchFilter, PlayerStatisticsChartFilter
from .teamresult import TeamResultsFilter, TeamResultsLeagueFilter



from .calendar_items import (
    CalendarItemRepresentationViewSet,
    CalendarItemViewSet,
    InfiniteCalendarItemViewSet,
    OwnCalendarItemViewSet,
    get_ics,
)
from .conference_room import (
    BuildingModelViewSet,
    BuildingRepresentationViewSet,
    ConferenceRoomModelViewSet,
    ConferenceRoomRepresentationViewSet,
)



from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
    create_simple_section,
)


def test_create_simple_display():
    assert create_simple_display([["field1", "field2"], ["field3", "field4"]]).serialize() == {
        "navigationType": "tab",
        "pages": [
            {
                "layouts": {
                    "default": {
                        "gridTemplateAreas": "'field1 field2' 'field3 field4'",
                        "inlines": [],
                        "sections": [],
                        "gap": "10px",
                        "gridAutoColumns": "1fr",
                        "gridAutoRows": "min-content",
                    }
                },
                "title": None,
            }
        ],
    }


def test_create_simple_section():
    section = create_simple_section(
        key="key", title="title", grid_template_areas=[["field1", "field2"]], inline_key="field1", collapsible=False
    ).serialize()
    assert section["key"] == "key"
    assert section["title"] == "title"
    assert section["display"]["pages"][0]["layouts"]["default"]["gridTemplateAreas"] == "'field1 field2'"  # type: ignore
    assert section["display"]["pages"][0]["layouts"]["default"]["inlines"] == (  # type: ignore
        [{"key": "field1", "endpoint": "field1", "hide_controls": False}]
    )
    assert not section["collapsible"]



from django.db.models import Case, Value, When
from dynamic_preferences.users.registries import user_preferences_registry
from dynamic_preferences.users.viewsets import UserPreferencesViewSet as BaseUserPreferencesViewSet


class UserPreferencesViewSet(BaseUserPreferencesViewSet):
    def get_queryset(self):
        obsolete_expressions = []
        weight_expressions = []
        # we mount a when statement dynamically from the static user preference registry
        # if the condition is matched, we assume the preference is not obsolete.
        for section_name, preferences in user_preferences_registry.items():
            for pref in preferences.values():
                obsolete_expressions.append(When(section=section_name, name=pref.name, then=Value(False)))
                weight_expressions.append(
                    When(section=section_name, name=pref.name, then=Value(getattr(pref, "weight", 0)))
                )
        return (
            super()
            .get_queryset()
            .annotate(
                is_obsolete=Case(*obsolete_expressions, default=Value(True)),
                weight=Case(*weight_expressions, default=Value(0)),
            )
            .exclude(is_obsolete=True)
            .order_by("weight", "name")
        )



from wbcore import viewsets
from wbcore.contrib.example_app.models import Season
from wbcore.contrib.example_app.serializers import (
    SeasonModelSerializer,
    SeasonRepresentationSerializer,
)
from wbcore.contrib.example_app.viewsets.displays import SeasonDisplayConfig


class SeasonModelViewSet(viewsets.ModelViewSet):
    display_config_class = SeasonDisplayConfig
    queryset = Season.objects.all()
    search_fields = ("name", "winner")
    serializer_class = SeasonModelSerializer
    ordering_fields = ("name", "winner", "top_scorer", "start_date", "end_date")
    ordering = ("name", "id")
    filterset_fields = ("name", "winner")


class SeasonRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = Season.objects.all()
    serializer_class = SeasonRepresentationSerializer
    search_fields = ("name",)



powers = [10**x for x in (3, 6, 9, 12, 15, 18, 21, 24, 27, 30, 33, 100)]

human_powers = (
    ("Thousand", "Thousand"),
    ("Million", "Million"),
    ("Billion", "Billion"),
    ("Trillion", "Trillion"),
    ("Quadrillion", "Quadrillion"),
    ("Quintillion", "Quintillion"),
    ("Sextillion", "Sextillion"),
    ("Septillion", "Septillion"),
    ("Octillion", "Octillion"),
    ("Nonillion", "Nonillion"),
    ("Decillion", "Decillion"),
    ("Googol", "Googol"),
)


def intword(value, format=".1f"):
    """Converts a large integer to a friendly text representation.
    Works best for numbers over 1 million. For example, 1_000_000 becomes "1.0 million",
    1200000 becomes "1.2 million" and "1_200_000_000" becomes "1.2 billion". Supports up
    to decillion (33 digits) and googol (100 digits).
    Examples:
        ```pycon
        >>> intword("100")
        '100'
        >>> intword("12400")
        '12.4 thousand'
        >>> intword("1000000")
        '1.0 million'
        >>> intword(1_200_000_000)
        '1.2 billion'
        >>> intword(8100000000000000000000000000000000)
        '8.1 decillion'
        >>> intword(None) is None
        True
        >>> intword("1234000", "%0.3f")
        '1.234 million'
        ```
    Args:
        value (int, float, str): Integer to convert.
        format (str): To change the number of decimal or general format of the number
            portion.
    Returns:
        str: Friendly text representation as a string, unless the value passed could not
        be coaxed into an `int`.
    """
    try:
        value = int(value)
    except (TypeError, ValueError):
        return value

    output_format = format
    format = "%" + format

    if value < powers[0]:
        return str(value)
    for ordinal, power in enumerate(powers[1:], 1):
        if value < power:
            chopped = value / float(powers[ordinal - 1])
            if float(format % chopped) == float(10**3):
                chopped = value / float(powers[ordinal])
                singular, plural = human_powers[ordinal]
                return f"{chopped:{output_format}} {singular}"
            else:
                singular, plural = human_powers[ordinal - 1]
                return f"{chopped:{output_format}} {singular}"
    return str(value)



import random

import factory
from django.core.files.base import ContentFile
from django.db.models import signals

from ..models import Company, CompanyType, CustomerStatus, Entry, Person, Specialization


class CustomerStatusFactory(factory.django.DjangoModelFactory):
    title = factory.Faker("text", max_nb_chars=32)

    class Meta:
        model = CustomerStatus


class CompanyTypeFactory(factory.django.DjangoModelFactory):
    title = factory.Faker("text", max_nb_chars=32)

    class Meta:
        model = CompanyType
        django_get_or_create = ["title"]


class BankCompanyTypeFactory(CompanyTypeFactory):
    title = "Bank"


class SpecializationFactory(factory.django.DjangoModelFactory):
    title = factory.Faker("text", max_nb_chars=32)

    class Meta:
        model = Specialization


@factory.django.mute_signals(signals.pre_save, signals.post_save)
class EntryBaseFactory(factory.django.DjangoModelFactory):
    class Meta:
        abstract = False


class EntryContactBaseFactory(factory.django.DjangoModelFactory):
    class Meta:
        abstract = False

    profile_image = factory.django.ImageField(filename="image_profile.jpeg")

    banking = factory.RelatedFactory("wbcore.contrib.directory.factories.contacts.BankingContactFactory", "entry")
    addresses = factory.RelatedFactory("wbcore.contrib.directory.factories.contacts.WebsiteContactFactory", "entry")


class EntryFactory(EntryBaseFactory):
    class Meta:
        model = Entry

    # entry_type
    signature = factory.django.ImageField(filename="signature.jpeg")
    activity_heat = random.uniform(0.0, 1.0)
    salutation = factory.Faker("text", max_nb_chars=64)

    @factory.post_generation
    def relationship_managers(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for in_charg in extracted:
                self.relationship_managers.add(in_charg)


class CompanyFactory(EntryFactory):
    class Meta:
        model = Company

    name = factory.Faker("company")
    customer_status = factory.SubFactory(CustomerStatusFactory)
    type = factory.SubFactory(CompanyTypeFactory)
    tier = factory.Iterator(Company.Tiering.choices, getter=lambda tier: tier[0])


class EntryContactFactory(EntryFactory):
    @factory.post_generation
    def telephones(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for telephone in extracted:
                self.telephones.add(telephone)

    @factory.post_generation
    def emails(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for email in extracted:
                self.emails.add(email)


class EmployersCompanyFactory(CompanyFactory):
    @factory.post_generation
    def employees(self, create, extracted, **kwargs):
        if extracted:
            for _ in range(extracted):
                PersonFactory(employers=(self,))


class BankFactory(CompanyFactory):
    type = factory.SubFactory(BankCompanyTypeFactory)


class PersonFactory(EntryFactory):
    class Meta:
        model = Person

    first_name = factory.Faker("first_name")
    last_name = factory.Faker("last_name")
    birthday = factory.Faker("date_object")
    description = factory.Faker("text")
    personality_profile_red = random.randint(1, 5)
    personality_profile_green = random.randint(1, 5)
    personality_profile_blue = random.randint(1, 5)
    prefix = factory.Iterator(Person.Prefix.choices, getter=lambda prefix: prefix[0])

    @factory.post_generation
    def employers(self, create, extracted, **kwargs):
        if extracted:
            for employer in extracted:
                self.employers.add(employer)


class PersonWithEmployerFactory(PersonFactory):
    employers = factory.RelatedFactory(
        "wbcore.contrib.directory.factories.relationships.EmployerEmployeeRelationshipFactory",
        factory_related_name="employee",
    )


class PersonSignatureFactory(PersonFactory):
    signature = factory.LazyAttribute(
        lambda _: ContentFile(
            factory.django.ImageField()._make_data({"width": 1024, "height": 768}),
            "example.jpg",
        )
    )


class CompanyWithEmployerEmployeeRelationshipFactory(CompanyFactory):
    employees = factory.RelatedFactory(
        "wbcore.contrib.directory.factories.relationships.EmployerEmployeeRelationshipFactory",
        factory_related_name="employer",
    )


class ClientFactory(PersonFactory):
    @factory.post_generation
    def clients(self, create, extracted, **kwargs):
        if extracted:
            for person in extracted:
                self.clients.add(person)


class RandomClientFactory(ClientFactory):
    @factory.post_generation
    def clients(self, create, extracted, **kwargs):
        self.clients.add(PersonFactory())


class UnemployedPersonFactory(PersonFactory):
    @factory.post_generation
    def employers(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            self.employers.set()



from django.contrib.contenttypes.models import ContentType
from django.utils.translation import gettext as _

from wbcore.contrib.workflow.models import Process, ProcessStep, Workflow


def check_workflow_for_instance(sender, instance, created, *args, **kwargs):
    # Fail all active process steps with statuses that don't match the instance's status
    active_process_steps = ProcessStep.objects.filter(
        state__in=[ProcessStep.StepState.ACTIVE, ProcessStep.StepState.WAITING],
        process__instance_id=instance.pk,
        process__content_type=ContentType.objects.get_for_model(instance),
        process__state=Process.ProcessState.ACTIVE,
    )
    for process_step in active_process_steps:
        if hasattr(instance, process_step.process.workflow.status_field) and not process_step.status == getattr(
            instance, process_step.process.workflow.status_field
        ):
            process_step.step.get_casted_step().set_failed(process_step, _("Invalid status detected!"))

    # Start workflow if instance's status matches the status of a start step
    start_steps = Workflow.get_start_steps_for_instance(instance)
    for start_step in start_steps:
        start_step.workflow.start_workflow(start_step, instance)



from __future__ import annotations

import datetime
from decimal import Decimal
from pathlib import Path

from celery import current_app, shared_task
from django.contrib.postgres.fields import DateRangeField
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models import Q
from django.dispatch import receiver
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from django_fsm import FSMField, transition
from psycopg.types.range import TimestamptzRange
from slugify import slugify

from wbcore.contrib.agenda.models import CalendarItem
from wbcore.contrib.authentication.models import User
from wbcore.contrib.color.enums import WBColor
from wbcore.contrib.color.models import ColorField
from wbcore.contrib.directory.models import Company, Person
from wbcore.contrib.geography.models import Geography
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.icons.models import IconField
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton
from wbcore.models.orderable import OrderableModel
from wbcore.utils.models import ActiveObjectManager, ComplexToStringMixin


def upload_to_profile_images(instance, filename):
    file_extension = "".join(Path(filename).suffixes)
    return f"example_app/sport_person/profile_images/{instance.id}{file_extension}"


class Role(models.Model):
    title = models.CharField(max_length=100, unique=True, verbose_name=_("Title"))
    slugify_title = models.CharField(max_length=100, unique=True, null=True, blank=True)

    sport_persons: models.QuerySet[SportPerson]

    def __str__(self) -> str:
        return self.title

    def save(self, *args, **kwargs):
        self.slugify_title = slugify(self.title, separator=" ")
        super().save(*args, **kwargs)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "example_app:role"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "example_app:rolerepresentation-list"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{title}}"

    class Meta:
        verbose_name = _("Role")
        verbose_name_plural = _("Roles")


class SportPerson(ComplexToStringMixin):
    roles = models.ManyToManyField(to=Role, blank=True, related_name="sport_persons", verbose_name=_("Roles"))
    first_name = models.CharField(max_length=255, verbose_name=_("First Name"))
    last_name = models.CharField(max_length=255, verbose_name=_("Last Name"))
    profile = models.OneToOneField(Person, on_delete=models.CASCADE, related_name="sport_profile")
    profile_image = models.ImageField(blank=True, null=True, upload_to=upload_to_profile_images, max_length=256)

    coached_team: Team
    commissioned_leagues: models.QuerySet[League]
    refereed_matches: models.QuerySet[Match]
    events: models.QuerySet[Event]

    def str_full(self) -> str:
        """
        Get the string representation including the coached team or commissioned league.
        """

        full_name = f"{self.first_name} {self.last_name}"

        # Why is it necessary to exclude the null values?
        if Team.objects.exclude(coach__isnull=True).filter(coach_id=self.id).exists():
            full_name += _(f" (Coach of {self.coached_team.name})")
        elif League.objects.exclude(commissioner__isnull=True).filter(commissioner_id=self.id).exists():
            full_name += _(f" (Commissioner of {self.commissioned_leagues.first().name})")

        return full_name

    def compute_str(self) -> str:
        self.slugify_computed_str = slugify(self.str_full(), separator=" ")
        return self.str_full()

    def get_player(self) -> Player | None:
        """Tries to cast a sport person into its child player class."""

        try:
            return Player.objects.get(id=self.pk)
        except Player.DoesNotExist:
            return None

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "example_app:person"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "example_app:personrepresentation-list"

    class Meta:
        verbose_name = _("Person")
        verbose_name_plural = _("Persons")


class Sport(models.Model):
    name = models.CharField(max_length=100, unique=True, verbose_name=_("Name"))
    slugify_name = models.CharField(max_length=100, unique=True, null=True, blank=True)
    rules = models.TextField(max_length=100, null=True, blank=True, verbose_name=_("Rules"))
    match_duration = models.PositiveIntegerField(verbose_name=_("Match Duration (min)"))

    matches: models.QuerySet[Match]
    event_types: models.QuerySet[EventType]
    leagues: models.QuerySet[League]

    def __str__(self) -> str:
        return self.name

    def save(self, *args, **kwargs):
        self.slugify_name = slugify(self.name, separator=" ")
        super().save(*args, **kwargs)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "example_app:sport"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "example_app:sportrepresentation-list"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"

    class Meta:
        verbose_name = _("Sport")
        verbose_name_plural = _("Sports")


@receiver(models.signals.post_save, sender=Sport)
def post_save_sport(sender, instance: Sport, created: bool, raw: bool, **kwargs):
    """Triggers the save signals of the sport's leagues updating their computed_str"""

    if not raw:
        if leagues := instance.leagues.all():
            for league in leagues:
                league.save()


class League(ComplexToStringMixin, models.Model):
    name = models.CharField(max_length=100, verbose_name=_("Name"))
    sport = models.ForeignKey(
        to=Sport,
        on_delete=models.PROTECT,
        related_name="leagues",
        verbose_name=_("Sport"),
    )
    country = models.ForeignKey(
        to=Geography,
        limit_choices_to={"level": 1},
        related_name="leagues",
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        verbose_name=_("Country"),
    )
    established_date = models.DateField(null=True, blank=True, verbose_name=_("Established Date"))
    commissioner = models.ForeignKey(
        to=SportPerson,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="commissioned_leagues",
        verbose_name=_("Commissioner"),
    )
    website = models.URLField(max_length=200, null=True, blank=True, verbose_name=_("Website"))
    points_per_win = models.PositiveIntegerField(verbose_name=_("Points Per Win"))
    points_per_draw = models.PositiveIntegerField(verbose_name=_("Points Per Draw"), default=1)
    points_per_loss = models.PositiveIntegerField(verbose_name=_("Points Per Loss"), default=0)
    teams = models.ManyToManyField(
        "Team",
        related_name="leagues",
        blank=True,
        through="TeamResults",
        through_fields=("league", "team"),
        verbose_name=_("Teams"),
    )

    matches: models.QuerySet[Match]
    results: models.QuerySet[TeamResults]
    seasons: models.QuerySet[Season]

    def __str__(self) -> str:
        return f"{self.name} ({self.sport.name})"

    def compute_str(self) -> str:
        return self.__str__()

    class Meta:
        verbose_name = _("League")
        verbose_name_plural = _("Leagues")
        constraints = [
            models.UniqueConstraint(fields=["name", "sport"], name="league_name_sport"),
        ]

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "example_app:league"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "example_app:leaguerepresentation-list"


@receiver(models.signals.post_save, sender=League)
def post_save_league(sender, instance: League, created: bool, raw: bool, **kwargs):
    """Triggers the save signals of the league's commissioner updating their computed_str"""

    if not raw:
        if instance.commissioner:
            instance.commissioner.save()


class TeamResults(models.Model):
    team = models.ForeignKey(
        on_delete=models.CASCADE,
        to="Team",
        verbose_name=_("Team"),
        related_name="results",
    )
    league = models.ForeignKey(
        on_delete=models.CASCADE,
        to=League,
        verbose_name=_("League"),
        related_name="results",
    )
    points = models.PositiveIntegerField(verbose_name=_("Points"), default=0)
    match_points_for = models.PositiveIntegerField(verbose_name=_("Match Points For"), default=0)
    match_points_against = models.PositiveIntegerField(verbose_name=_("Match Points Against"), default=0)
    match_point_difference = models.IntegerField(verbose_name=_("Match Point Difference"), default=0)
    wins = models.PositiveIntegerField(verbose_name=_("Wins"), default=0)
    draws = models.PositiveIntegerField(verbose_name=_("Draws"), default=0)
    losses = models.PositiveIntegerField(verbose_name=_("Losses"), default=0)
    form = models.CharField(verbose_name=_("Form"), max_length=5, default="-----")

    def __str__(self) -> str:
        return _("{} in {}").format(self.team.name, self.league.name)

    def save(self, *args, **kwargs):
        self.match_point_difference = self.match_points_for - self.match_points_against
        return super().save(*args, **kwargs)

    class Meta:
        verbose_name = _("Team Result")
        verbose_name_plural = _("Team Results")
        constraints = [
            models.UniqueConstraint(fields=["team", "league"], name="unique_team_league"),
        ]

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "example_app:teamresults"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "example_app:teamresultsrepresentation-list"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{team__name}} in {{league__computed_str}}"


class Stadium(models.Model):
    name = models.CharField(max_length=100, unique=True, verbose_name=_("Name"))
    slugify_name = models.CharField(max_length=100, unique=True, null=True, blank=True)
    city = models.ForeignKey(
        to=Geography,
        limit_choices_to={"level": 3},
        related_name="stadiums",
        on_delete=models.PROTECT,
        verbose_name=_("City"),
        null=True,
        blank=True,
    )
    standing_capacity = models.PositiveIntegerField(default=0, verbose_name=_("Standing Capacity"))
    seating_capacity = models.PositiveIntegerField(default=0, verbose_name=_("Seating Capacity"))
    guest_rating = models.IntegerField(
        validators=[MinValueValidator(0), MaxValueValidator(4)],
        default=3,
        verbose_name=_("Guest Rating"),
    )

    teams_playing: models.QuerySet[Team]
    matches: models.QuerySet[Match]

    def __str__(self) -> str:
        return self.name

    def save(self, *args, **kwargs):
        self.slugify_name = slugify(self.name, separator=" ")
        super().save(*args, **kwargs)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "example_app:stadium"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "example_app:stadiumrepresentation-list"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"

    @property
    def total_capacity(self):
        return self.standing_capacity + self.seating_capacity

    class Meta:
        verbose_name = _("Stadium")
        verbose_name_plural = _("Stadiums")


class Match(ComplexToStringMixin, CalendarItem):
    class MatchStatus(models.TextChoices):
        SCHEDULED = "SCHEDULED", _("Scheduled")
        ONGOING = "ONGOING", _("Ongoing")
        FINISHED = "FINISHED", _("Finished")

        @classmethod
        def get_color_map(cls) -> list[tuple[Match.MatchStatus, str]]:
            colors = [
                WBColor.YELLOW_LIGHT.value,
                WBColor.BLUE_LIGHT.value,
                WBColor.GREEN_LIGHT.value,
            ]
            return [status for status in zip(cls, colors, strict=False)]

    home = models.ForeignKey(
        to="Team",
        related_name="home_matches",
        on_delete=models.CASCADE,
        verbose_name=_("Home"),
    )
    away = models.ForeignKey(
        to="Team",
        related_name="away_matches",
        on_delete=models.CASCADE,
        verbose_name=_("Away"),
    )
    date_time = models.DateTimeField(verbose_name=_("Date Time"))
    stadium = models.ForeignKey(
        to=Stadium,
        related_name="matches",
        on_delete=models.CASCADE,
        verbose_name=_("Stadium"),
    )
    status = FSMField(
        choices=MatchStatus.choices,
        verbose_name=_("Status"),
        default=MatchStatus.SCHEDULED,
    )
    score_home = models.PositiveIntegerField(verbose_name=_("Home Score"), default=0, editable=False)
    score_away = models.PositiveIntegerField(verbose_name=_("Away Score"), default=0, editable=False)
    referee = models.ForeignKey(
        to=SportPerson,
        null=True,
        blank=True,
        related_name="refereed_matches",
        on_delete=models.SET_NULL,
    )
    league = models.ForeignKey(
        to=League,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="matches",
        verbose_name=_("League"),
    )
    winner = models.ForeignKey(
        "Team",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="won_matches",
    )
    sport = models.ForeignKey(
        to=Sport,
        on_delete=models.PROTECT,
        related_name="matches",
        verbose_name=_("Sport"),
    )
    task_id = models.CharField(blank=True, editable=False, max_length=50)

    events: models.QuerySet[Event]

    def has_permissions(self: Match, user: User) -> bool:
        if user.is_superuser or user.has_perm("wbcore.change_match_status"):
            return True
        return False

    @transition(
        field=status,
        source=[MatchStatus.SCHEDULED],
        target=MatchStatus.ONGOING,
        permission=has_permissions,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("example_app:match",),
                icon=WBIcon.START.icon,
                key="start",
                label=_("Start"),
                action_label=_("Starting"),
                description_fields=_("Are you sure you want to start the match?"),
            )
        },
    )
    def start(self, by=None, description=None, **kwargs):
        pass

    @transition(
        field=status,
        source=[MatchStatus.ONGOING],
        target=MatchStatus.FINISHED,
        permission=has_permissions,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("example_app:match",),
                icon=WBIcon.TIME_UP.icon,
                key="end",
                label=_("End"),
                action_label=_("Ending"),
                description_fields=_("Are you sure you want to end the match?"),
            )
        },
    )
    def end(self, by=None, description=None, **kwargs):
        pass

    def __str__(self) -> str:
        return f"{self.home.name} vs. {self.away.name}"

    def recompute_computed_str(self):
        """A separate function to only update the computed_str to not have to call the compute-heavy save method each time"""

        Match.objects.filter(id=self.pk).update(computed_str=self.compute_str())

    def reschedule_task(self):
        """Reschedules the task that automatically starts a match at the given date time"""

        if task_id := self.task_id:
            current_app.control.revoke(task_id, terminate=True)
        new_task = start_match.apply_async(eta=self.date_time, args=[self.pk])
        self.task_id = new_task.id

    def get_color(self) -> str:
        return "#f7be2f"  # dark yellow

    def get_icon(self) -> str:
        return WBIcon.GOAL.icon

    def save(self, *args, **kwargs):
        start_time = self.date_time
        end_time = start_time + datetime.timedelta(minutes=self.sport.match_duration)
        self.period = TimestamptzRange(start_time, end_time)
        self.title = self.__str__()

        if self.status == self.MatchStatus.FINISHED and self.score_home > self.score_away:
            self.winner = self.home
        elif self.status == self.MatchStatus.FINISHED and self.score_away > self.score_home:
            self.winner = self.away
        else:
            self.winner = None

        # if we change the date time we need to reschedule the task that automatically starts the match
        try:
            old_instance = Match.objects.get(pk=self.pk)
            if old_instance.date_time != self.date_time:
                self.reschedule_task()
        except Match.DoesNotExist:
            self.reschedule_task()

        if self.league and self.task_id and self.status == self.MatchStatus.FINISHED:
            home_results, created1 = TeamResults.objects.get_or_create(team=self.home, league=self.league)
            away_results, created2 = TeamResults.objects.get_or_create(team=self.away, league=self.league)

            # Update points, W/L/D & form
            # Home win
            if self.score_home > self.score_away:
                home_results.wins += 1
                home_results.form = home_results.form[1:] + "W"
                home_results.points += self.league.points_per_win
                away_results.losses += 1
                away_results.form = away_results.form[1:] + "L"
                away_results.points += self.league.points_per_loss

            # Away win
            elif self.score_home < self.score_away:
                home_results.losses += 1
                home_results.form = home_results.form[1:] + "L"
                home_results.points += self.league.points_per_loss
                away_results.wins += 1
                away_results.form = away_results.form[1:] + "W"
                away_results.points += self.league.points_per_win

            # Draw
            else:
                home_results.draws += 1
                home_results.form = home_results.form[1:] + "D"
                home_results.points += self.league.points_per_draw
                away_results.draws += 1
                away_results.form = away_results.form[1:] + "D"
                away_results.points += self.league.points_per_draw

            # Update match points
            home_results.match_points_for += self.score_home
            home_results.match_points_against += self.score_away
            away_results.match_points_for += self.score_away
            away_results.match_points_against += self.score_home

            # Update match point difference
            home_results.save()
            away_results.save()

            # Set empty task ID to indicate that this match's result has been synchronized to the league table
            self.task_id = ""

        return super().save(*args, **kwargs)

    def compute_str(self) -> str:
        return self.__str__()

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "example_app:match"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "example_app:matchrepresentation-list"

    class Meta:
        verbose_name = _("Match")
        verbose_name_plural = _("Matches")
        constraints = [
            models.UniqueConstraint(fields=["home", "away", "date_time"], name="match_home_away_date_time"),
            models.CheckConstraint(check=~models.Q(home=models.F("away")), name="check_match_home_away"),
        ]
        permissions = [("change_match_status", "Change Match Status")]


@shared_task
def start_match(match_id: int):
    """Sets the match status from scheduled to ongoing.

    Args:
        match_id (int): ID of the match in question
    """

    match = Match.objects.get(pk=match_id)
    if match.status == Match.MatchStatus.SCHEDULED:
        match.status = Match.MatchStatus.ONGOING
        match.save()


@receiver(models.signals.post_save, sender=Match)
def update_team_duration_since_last_win(sender, instance: Match, **kwargs):
    if instance.winner:
        team = instance.winner

        # Calculate the duration since the current match win
        now = timezone.now()
        match_date = instance.date_time
        new_duration_since_win = now - match_date

        # Update only if this win is more recent (i.e., duration is shorter)
        if team.duration_since_last_win is None or new_duration_since_win < team.duration_since_last_win:
            team.duration_since_last_win = new_duration_since_win
            team.save()


@receiver(models.signals.post_save, sender=Match)
def post_save_match(sender, instance: Match, created: bool, raw: bool, **kwargs):
    """Sets the entities in the parent calendar item"""

    if raw:
        return

    home_sport_persons = SportPerson.objects.filter(
        Q(id__in=instance.home.current_players.all()) | Q(coached_team=instance.home)
    )
    away_sport_persons = SportPerson.objects.filter(
        Q(id__in=instance.away.current_players.all()) | Q(coached_team=instance.away)
    )
    referee = SportPerson.objects.none()
    if instance.referee:
        referee = SportPerson.objects.filter(id=instance.referee.id)
    match_persons_ids = home_sport_persons.union(away_sport_persons, referee).values_list("profile__id", flat=True)

    instance.entities.set(match_persons_ids)
    instance.entities.add(instance.home, instance.away)


class Team(OrderableModel, Company):
    slugify_name = models.CharField(max_length=100, unique=True, null=True, blank=True)
    city = models.ForeignKey(
        to=Geography,
        limit_choices_to={"level": 3},
        related_name="teams",
        on_delete=models.PROTECT,
        verbose_name=_("City"),
        null=True,
        blank=True,
    )
    founded_date = models.DateField(null=True, blank=True, verbose_name=_("Founded Date"))
    coach = models.OneToOneField(
        to=SportPerson,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="coached_team",
        verbose_name=_("Coach"),
    )
    home_stadium = models.ForeignKey(
        to=Stadium,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="teams_playing",
        verbose_name=_("Home Stadium"),
    )
    opponents = models.ManyToManyField(
        to="self",
        blank=True,
        through=Match,
        through_fields=("home", "away"),
        verbose_name=_("Opponents"),
    )
    duration_since_last_win = models.DurationField(
        blank=True,
        null=True,
    )
    email = models.EmailField(verbose_name=_("Email Address"), null=True, blank=True)
    phone_number = models.CharField(max_length=255, verbose_name=_("Phone Number"), null=True, blank=True)
    objects = ActiveObjectManager()
    website = models.URLField(default="https://duckduckgo.com/", verbose_name=_("Website"))
    current_players: models.QuerySet[Player]
    former_players: models.QuerySet[Player]
    home_matches: models.QuerySet[Match]
    away_matches: models.QuerySet[Match]
    results: models.QuerySet[TeamResults]
    won_seasons: models.QuerySet[Season]
    won_matches: models.QuerySet[Match]

    def __str__(self) -> str:
        return self.name

    def save(self, *args, **kwargs):
        self.slugify_name = slugify(self.name, separator=" ")
        super().save(*args, **kwargs)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "example_app:team"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "example_app:teamrepresentation-list"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"

    class Meta(OrderableModel.Meta):
        verbose_name = _("Team")
        verbose_name_plural = _("Teams")


@receiver(models.signals.post_save, sender=Team)
def post_save_team(sender, instance: Team, created: bool, raw: bool, **kwargs):
    """Triggers the save signals of the team's coaches, current players and matches updating their computed_str"""

    if not raw:
        if instance.coach:
            instance.coach.save()
        if players := instance.current_players.all():
            for player in players:
                player.save()
        if home_matches := instance.home_matches.all():
            for home_match in home_matches:
                home_match.recompute_computed_str()
        if away_matches := instance.away_matches.all():
            for away_match in away_matches:
                away_match.recompute_computed_str()


class Player(OrderableModel, SportPerson):  # noqa
    PARTITION_BY = PARENT_FK = "current_team"

    position = models.CharField(max_length=50, null=True, blank=True, verbose_name=_("Position"))
    current_team = models.ForeignKey(
        to=Team,
        related_name="current_players",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        verbose_name=_("Current Team"),
    )
    former_teams = models.ManyToManyField(
        to=Team,
        related_name="former_players",
        blank=True,
        verbose_name=_("Former Teams"),
    )
    transfer_value = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        default=Decimal(0),
        verbose_name=_("Market Value"),
    )
    is_active = models.BooleanField(default=True, verbose_name=_("Is Active"))
    is_injured = models.BooleanField(default=False, verbose_name=_("Is Injured"))
    player_strength = models.IntegerField(
        validators=[MinValueValidator(0), MaxValueValidator(5)],
        default=3,
        verbose_name=_("Player Strength"),
    )
    game_activity = models.FloatField(
        validators=[MinValueValidator(0), MaxValueValidator(1)],
        default=0,
        verbose_name=_("Game Activity"),
    )
    top_scorer_seasons: models.QuerySet[Season]

    def str_full(self) -> str:
        """
        Get the string representation including the position and the current team.
        """

        full_name = f"{self.first_name} {self.last_name}"

        if self.current_team and self.position:
            full_name += _(f" ({self.position} at {self.current_team.name})")
        elif self.current_team:
            full_name += _(f" ({self.current_team.name})")

        return full_name

    def compute_str(self) -> str:
        self.slugify_computed_str = slugify(self.str_full(), separator=" ")
        return self.str_full()

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "example_app:player"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "example_app:playerrepresentation-list"

    class Meta(OrderableModel.Meta):
        verbose_name = _("Player")
        verbose_name_plural = _("Players")


class Event(models.Model):
    event_description = models.TextField(verbose_name=_("Event Description"), max_length=512, blank=True, null=True)
    person = models.ForeignKey(
        to=SportPerson,
        related_name="events",
        on_delete=models.CASCADE,
        verbose_name=_("Person"),
    )
    match = models.ForeignKey(
        to=Match,
        related_name="events",
        on_delete=models.CASCADE,
        verbose_name=_("Match"),
    )
    minute = models.PositiveIntegerField(verbose_name=_("Minute"))
    event_type = models.ForeignKey(
        to="EventType",
        related_name="events",
        on_delete=models.CASCADE,
        verbose_name=_("Event Type"),
    )

    def __str__(self) -> str:
        return f"{self.event_type.name} ({self.minute}.) - {self.person.computed_str}"

    def save(self, *args, **kwargs):
        match = self.match
        if player := self.person.get_player():
            if player in match.home.current_players.all():
                match.score_home += self.event_type.points
                match.save()
            elif player in match.away.current_players.all():
                match.score_away += self.event_type.points
                match.save()
        super().save(*args, **kwargs)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "example_app:event"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "example_app:eventrepresentation-list"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{event_type__name}} ({{minute}}.) - {{person}}"

    class Meta:
        verbose_name = _("Event")
        verbose_name_plural = _("Events")
        constraints = [
            models.UniqueConstraint(
                fields=["person", "match", "minute", "event_type"],
                name="event_person_match_minute_event_type",
            ),
        ]


class EventType(models.Model):
    name = models.CharField(max_length=100, verbose_name=_("Name"))
    points = models.PositiveIntegerField(
        verbose_name=_("Points"),
        help_text=_("Number of points awarded to a player's team per event"),
        default=1,
    )
    sport = models.ForeignKey(
        to=Sport,
        related_name="event_types",
        on_delete=models.CASCADE,
        verbose_name=_("Sport"),
    )
    icon = IconField(
        max_length=128,
        verbose_name=_("Icon"),
        default=WBIcon.GOAL.icon,
    )
    color = ColorField(null=True, blank=True)

    events: models.QuerySet[Event]

    def __str__(self) -> str:
        return self.name

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "example_app:eventtype"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "example_app:eventtyperepresentation-list"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"

    class Meta:
        verbose_name = _("Event Type")
        verbose_name_plural = _("Event Types")
        constraints = [
            models.UniqueConstraint(fields=["name", "sport"], name="event_type_name_sport"),
        ]


class Season(models.Model):
    league = models.ForeignKey("League", on_delete=models.CASCADE, related_name="seasons")
    date_range = DateRangeField()
    name = models.CharField(max_length=255, blank=True)
    file = models.FileField(upload_to="season_files/", blank=True, null=True)
    winner = models.ForeignKey(
        "Team",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="won_seasons",
    )
    top_scorer = models.ForeignKey(
        "Player",
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="top_scorer_seasons",
    )

    def __str__(self):
        return f"{self.name} ({self.league})"

    def save(self, *args, **kwargs):
        if self.date_range:
            start_date = self.date_range.lower
            end_date = self.date_range.upper

            if not self.name and (start_date and end_date):
                self.name = f"{self.league.name} / {start_date.year}-{end_date.year} Season"

        super().save(*args, **kwargs)

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "example_app:season"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "example_app:seasonrepresentation-list"

    class Meta:
        verbose_name = _("Season")
        verbose_name_plural = _("Seasons")



from __future__ import annotations

from datetime import date as date_lib
from decimal import Decimal

from django.db import models
from django.db.models import Case, CharField, Expression, ExpressionWrapper, F, Q, Subquery, Value, When
from django.db.models.functions import Concat
from persisting_theory import QuerySet

from wbcore.contrib.io.mixins import ImportMixin
from wbcore.models import WBModel
from wbcore.utils.models import LabelKeyMixin

from .import_export.handlers import CurrencyFXRatesImportHandler, CurrencyImportHandler


class CurrencyDefaultManager(models.Manager):
    def get_queryset(self) -> QuerySet:
        return (
            super()
            .get_queryset()
            .annotate(
                name_repr=Case(
                    When(Q(symbol="") | Q(symbol__isnull=True), then=F("title")),
                    default=Concat(F("title"), Value(" ("), F("symbol"), Value(")")),
                    output_field=CharField(),
                )
            )
        )


class Currency(ImportMixin, LabelKeyMixin, WBModel):
    import_export_handler_class = CurrencyImportHandler

    class Meta:
        verbose_name = "Currency"
        verbose_name_plural = "Currencies"
        ordering = ("title",)

    def __str__(self) -> str:
        if self.symbol:
            return f"{self.title} ({self.symbol})"
        return self.title

    def __repr__(self) -> str:
        return self.key

    title = models.CharField(max_length=255)
    symbol = models.CharField(max_length=10, blank=True, null=True)
    key = models.CharField(max_length=3, unique=True)

    objects = CurrencyDefaultManager()

    LABEL_KEY = "{{key}} ({{symbol}})"

    def convert(self, valuation_date: date_lib, other_currency: Currency, exact_lookup: bool = False) -> Decimal:
        """
        Change the base currency of one of its related fx rates

        Arguments:
            valuation_date {datetime.date} -- The convertion date
            other_currency {Currency} -- The convertion currency
        Returns:
            float -- The fx rate value with the new base currency
        """
        if self == other_currency:
            return Decimal(1.0)
        try:
            if exact_lookup:
                base = CurrencyFXRates.objects.get(date=valuation_date, currency=self)
                other = CurrencyFXRates.objects.get(date=valuation_date, currency=other_currency)
            else:
                base = (
                    CurrencyFXRates.objects.filter(date__lte=valuation_date, currency=self).order_by("-date").first()
                )
                other = (
                    CurrencyFXRates.objects.filter(date__lte=valuation_date, currency=other_currency)
                    .order_by("-date")
                    .first()
                )
                if not base or not other:
                    raise CurrencyFXRates.DoesNotExist
            return (1 / base.value) * other.value
        except CurrencyFXRates.DoesNotExist:
            if exact_lookup:
                return self.convert(valuation_date, other_currency, exact_lookup=False)
            raise CurrencyFXRates.DoesNotExist from None

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcore:currency:currency"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcore:currency:currencyrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name_repr}}"


class CurrencyFXRates(ImportMixin, models.Model):
    import_export_handler_class = CurrencyFXRatesImportHandler

    date = models.DateField()
    currency = models.ForeignKey(to="Currency", related_name="fx_rates", on_delete=models.CASCADE)
    value = models.DecimalField(max_digits=20, decimal_places=6)

    class Meta:
        verbose_name = "FX Rate"
        verbose_name_plural = "FX Rates"
        unique_together = ("date", "currency")
        indexes = [
            models.Index(
                name="currency_fx_rate_idx",
                fields=["date", "currency"],
            )
        ]

    def __str__(self) -> str:
        return f"{self.date:%d.%m.%Y} {self.currency.key}"

    @classmethod
    def get_fx_rates_subquery(
        cls,
        val_date: date_lib | str,
        currency: str = "product__currency",
        lookup_expr: str = "lte",
    ) -> Expression:
        """
        Create a subquery to annotate the rate at a given date based on currency__key

        Arguments:
            date {datetime.date|str} -- The fx date
            is_outerref_key {bool} -- Determines the type of date variable, if true is outerref key string
            currency_key {str} -- The Outerref field for currency__key
        Returns:
            Subquery -- Subquery for rates at a certain date
        """
        fx_rates = (
            cls.objects.filter(
                **{f"date__{lookup_expr}": models.OuterRef(val_date) if isinstance(val_date, str) else val_date},
                currency=models.OuterRef(currency),
            )
            .order_by("-date")
            .values("value")[:1]
        )

        return ExpressionWrapper(1 / Subquery(fx_rates), output_field=models.DecimalField())

    @classmethod
    def get_fx_rates_subquery_for_two_currencies(
        cls,
        date_key: str,
        start_currency: str | Currency,
        target_currency: str | Currency,
    ) -> Expression:
        """
        Create a subquery to returns the convertion rate from a start to a target currency

        Arguments:
            date_key {datetime.date} -- The convertion date
            start_currency {str} -- The Outerref field for start currency
            target_currency {str} -- The Outerref field for target currency
        Returns:
            Subquery -- Subquery for the convertion rate rates at a certain date
        """
        fx_rates_start = (
            CurrencyFXRates.objects.filter(
                date=models.OuterRef(date_key),
                currency=models.OuterRef(start_currency) if isinstance(start_currency, str) else start_currency,
            )
            .order_by("-date")
            .values("value")[:1]
        )
        fx_rates_target = (
            CurrencyFXRates.objects.filter(
                date=models.OuterRef(date_key),
                currency=models.OuterRef(target_currency) if isinstance(target_currency, str) else target_currency,
            )
            .order_by("-date")
            .values("value")[:1]
        )
        return ExpressionWrapper(
            Subquery(fx_rates_target) / Subquery(fx_rates_start), output_field=models.DecimalField()
        )

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{value}} ({{currency__symbol}})"



from decimal import Decimal

import pytest
from rest_framework.exceptions import ValidationError

from wbcore.serializers import DecimalField, FloatField, IntegerField
from wbcore.serializers.fields import YearField
from wbcore.serializers.fields.types import WBCoreType


class TestIntegerField:
    def setup_method(self):
        self.field = IntegerField()

    def test_not_none(self):
        assert self.field is not None

    @pytest.mark.parametrize("input, expected", [(1, 1), ("1", 1), (-1, -1), (0, 0)])
    def test_to_internal_value(self, input, expected):
        assert self.field.to_internal_value(input) == expected

    @pytest.mark.parametrize("input", [None, [], "a", "", {}, 1.2])
    def test_to_internal_value_validation_error(self, input):
        with pytest.raises(ValidationError):
            self.field.to_internal_value(input)

    def test_field_type(self):
        assert self.field.field_type == WBCoreType.NUMBER.value

    def test_representation(self):
        assert self.field.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": None,
                "type": self.field.field_type,
                "required": True,
                "read_only": False,
                "display_mode": "decimal",
                "precision": 0,
                "max_digits": 34,
                "decorators": [],
                "depends_on": [],
                "signed": True,
                "disable_formatting": False,
            },
        )


class TestDecimalField:
    def setup_method(self):
        self.field = DecimalField(decimal_places=2, max_digits=5)

    def test_not_none(self):
        assert self.field is not None

    @pytest.mark.parametrize(
        "input, expected",
        [(1.0, Decimal(1)), ("1.0", Decimal(1)), (-1.0, Decimal(-1)), (0, Decimal(0))],
    )
    def test_to_internal_value(self, input, expected):
        assert self.field.to_internal_value(input) == expected

    @pytest.mark.parametrize("input", [None, [], "a", "", {}])
    def test_to_internal_value_validation_error(self, input):
        with pytest.raises(ValidationError):
            self.field.to_internal_value(input)

    def test_field_type(self):
        assert self.field.field_type == WBCoreType.NUMBER.value

    def test_representation(self):
        assert self.field.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": None,
                "type": self.field.field_type,
                "required": True,
                "read_only": False,
                "precision": 2,
                "max_digits": 5,
                "decorators": [],
                "depends_on": [],
                "display_mode": "decimal",
                "signed": True,
                "disable_formatting": False,
            },
        )

    def test_percent_representation(self):
        _field = self.field
        _field.percent = True
        assert _field.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": None,
                "type": WBCoreType.PERCENT.value,
                "required": True,
                "read_only": False,
                "precision": 0,
                "max_digits": 5,
                "decorators": [],
                "depends_on": [],
                "display_mode": "decimal",
                "signed": True,
                "disable_formatting": False,
            },
        )


class TestFloatField:
    def setup_method(self):
        self.field = FloatField()

    def test_not_none(self):
        assert self.field is not None

    @pytest.mark.parametrize("input, expected", [(1.0, 1.0), ("1.0", 1.0), (-1.0, -1.0), (0, 0)])
    def test_to_internal_value(self, input, expected):
        assert self.field.to_internal_value(input) == expected

    @pytest.mark.parametrize("input", [None, [], "a", "", {}])
    def test_to_internal_value_validation_error(self, input):
        with pytest.raises(ValidationError):
            self.field.to_internal_value(input)

    def test_field_type(self):
        assert self.field.field_type == WBCoreType.NUMBER.value

    def test_representation(self):
        assert self.field.get_representation(None, "field_name") == (
            "field_name",
            {
                "key": "field_name",
                "label": None,
                "type": self.field.field_type,
                "required": True,
                "read_only": False,
                "precision": 2,
                "max_digits": 34,
                "decorators": [],
                "depends_on": [],
                "display_mode": "decimal",
                "signed": True,
                "disable_formatting": False,
            },
        )


class TestYearField:
    @pytest.mark.parametrize(
        (
            "key",
            "label",
        ),
        [
            ("Foo", "Bar"),
        ],
    )
    def test_year_field_values(self, key, label):
        field = YearField(label=label, precision=2)
        representation = field.get_representation(None, key)[1]
        assert representation["precision"] == 0
        assert representation["disable_formatting"]



from django.contrib import admin
from django.utils.html import format_html

from .models import ReleaseNote


@admin.register(ReleaseNote)
class ReleaseNoteModelAdmin(admin.ModelAdmin):
    list_display = (
        "version",
        "release_date",
        "summary",
        "module",
    )
    list_filter = ("module", "release_date")

    fields = (("version", "module", "release_date"), "summary", "rendered_notes")

    @staticmethod
    def rendered_notes(obj):
        return format_html(obj.notes)

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False



from django.db.models import Q, QuerySet
from django.utils.translation import gettext_lazy as _

from wbcore import filters
from wbcore.contrib.example_app.models import Match, Player, Role, SportPerson, Team


class SportPersonRepresentationFilter(filters.FilterSet):
    match = filters.ModelChoiceFilter(
        label=_("Participating In"),
        queryset=Match.objects.all(),
        endpoint=Match.get_representation_endpoint(),
        value_key=Match.get_representation_value_key(),
        label_key=Match.get_representation_label_key(),
        method="filter_by_match",
        help_text=_("All persons participating in a selected Match."),
    )

    def filter_by_match(self, queryset: QuerySet[SportPerson], name, value: Match) -> QuerySet[SportPerson]:
        return queryset.filter(
            Q(coached_team__home_matches=value)
            | Q(coached_team__away_matches=value)
            | Q(refereed_matches=value)
            | Q(
                id__in=Player.objects.filter(
                    Q(current_team__home_matches=value) | Q(current_team__away_matches=value)
                ).values("id")
            )
        ).distinct()

    class Meta:
        model = SportPerson
        fields = {"roles": ["exact", "icontains"]}


class SportPersonFilter(SportPersonRepresentationFilter):
    roles = filters.ModelMultipleChoiceFilter(
        label=_("Roles"),
        queryset=Role.objects.all(),
        endpoint=Role.get_representation_endpoint(),
        value_key=Role.get_representation_value_key(),
        label_key=Role.get_representation_label_key(),
    )

    class Meta:
        model = SportPerson
        fields = {
            "last_name": ["exact", "icontains"],
            "first_name": ["exact", "icontains"],
        }


class PlayerTeamFilter(SportPersonFilter):
    former_teams = filters.ModelMultipleChoiceFilter(
        label=_("Former Teams"),
        queryset=Team.objects.all(),
        endpoint=Team.get_representation_endpoint(),
        value_key=Team.get_representation_value_key(),
        label_key=Team.get_representation_label_key(),
    )
    transfer_value__gte = filters.NumberFilter(label=_("Market Value"), lookup_expr="gte", field_name="transfer_value")
    transfer_value__lte = filters.NumberFilter(label=_("Market Value"), lookup_expr="lte", field_name="transfer_value")

    class Meta(SportPersonFilter.Meta):
        model = Player
        fields = SportPersonFilter.Meta.fields | {
            "position": ["exact", "icontains"],
        }


class PlayerFilter(PlayerTeamFilter):
    current_team = filters.ModelMultipleChoiceFilter(
        label=_("Current Teams"),
        queryset=Team.objects.all(),
        endpoint=Team.get_representation_endpoint(),
        value_key=Team.get_representation_value_key(),
        label_key=Team.get_representation_label_key(),
    )



from contextlib import suppress

import django_filters
from django.contrib.postgres.fields import RangeField
from django_filters.constants import EMPTY_VALUES
from django_filters.utils import get_model_field

from wbcore.filters.mixins import WBCoreFilterMixin
from wbcore.forms import DateRangeField, DateTimeRangeField
from wbcore.utils.date import financial_performance_shortcuts
from wbcore.utils.date_builder.components import Component


class TimeFilter(WBCoreFilterMixin, django_filters.TimeFilter):
    filter_type = "time"


class DateTimeFilter(WBCoreFilterMixin, django_filters.DateTimeFilter):
    filter_type = "datetime"


class DateFilter(WBCoreFilterMixin, django_filters.DateFilter):
    filter_type = "date"


class ShortcutAndPerformanceMixin(WBCoreFilterMixin):
    def __init__(self, shortcuts: list | None = None, performance_mode: bool = False, *args, **kwargs):
        self.shortcuts = shortcuts
        self.performance_mode = performance_mode
        super().__init__(*args, **kwargs)

    def get_representation(self, request, name, view):
        representation, lookup_expr = super().get_representation(request, name, view)
        lookup_expr["input_properties"]["performance_mode"] = self.performance_mode

        if self.shortcuts:
            lookup_expr["input_properties"]["shortcuts"] = self.shortcuts

        return representation, lookup_expr


class DateRangeFilter(ShortcutAndPerformanceMixin, django_filters.Filter):
    field_class = DateRangeField
    filter_type = "daterange"
    initial_format = "%Y-%m-%d"

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("lookup_expr", "overlap")
        super().__init__(*args, **kwargs)

    @property
    def is_range(self) -> bool:
        if hasattr(self, "model"):
            field = get_model_field(self.model, self.field_name)
            return issubclass(field.__class__, RangeField)
        return False

    def _get_initial(self, *args):
        initial = super()._get_initial(*args)
        if initial is not None:
            lower = upper = None
            if isinstance(initial, tuple):
                lower, upper = initial

                # if the initial is a tuple of components, we need to convert them to string
                if isinstance(lower, Component) and isinstance(upper, Component):
                    return f"{lower},{upper}"

            elif hasattr(initial, "lower") and hasattr(initial, "upper"):
                lower, upper = initial.lower, initial.upper
            initial = f'{lower.strftime(self.initial_format) if lower else ""},{upper.strftime(self.initial_format) if upper else ""}'
        return initial

    def get_representation(self, request, name, view):
        representation, lookup_expr = super().get_representation(request, name, view)
        representation["lookup_expr"] = {"exact": self.field_name}
        with suppress(KeyError):  # TODO frontend needs to support both exact and overlaps lookup
            initial = representation["initial"].pop(self.lookup_expr)
            lookup_expr["input_properties"]["initial"]["exact"] = initial

        return representation, lookup_expr

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs
        if value:
            lower, upper = value.lower, value.upper
            filters = {}
            is_field_range = self.is_range
            if lower:
                if is_field_range:
                    filters[f"{self.field_name}__startswith__gte"] = lower
                else:
                    filters[f"{self.field_name}__gte"] = lower

            if upper:
                if is_field_range:
                    filters[f"{self.field_name}__endswith__lte"] = upper
                else:
                    filters[f"{self.field_name}__lte"] = upper

            if self.exclude:
                qs = qs.exclude(**filters)
            else:
                qs = qs.filter(**filters)
        return qs


class FinancialPerformanceDateRangeFilter(DateRangeFilter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, performance_mode=True, shortcuts=financial_performance_shortcuts, **kwargs)


class DateTimeRangeFilter(DateRangeFilter):
    field_class = DateTimeRangeField
    initial_format = "%Y-%m-%dT%H:%M:%S%z"
    filter_type = "datetimerange"



from configurations import values


class Cache:
    CACHE_REDIS_URL = values.Value(None, environ_prefix=None)
    CACHE_TIMEOUT = values.IntegerValue(
        10 * 24 * 3600, environ_prefix=None
    )  # The default (10 days) timeout in seconds

    @property
    def CACHES(self):  # noqa
        if self.CACHE_REDIS_URL:
            return {
                "default": {
                    "BACKEND": "django.core.cache.backends.redis.RedisCache",
                    "LOCATION": self.CACHE_REDIS_URL,
                    "TIMEOUT": self.CACHE_TIMEOUT,
                }
            }
        return {"default": {"BACKEND": "django.core.cache.backends.locmem.LocMemCache", "TIMEOUT": self.CACHE_TIMEOUT}}



from django.utils.translation import gettext as _
from rest_framework.reverse import reverse

from wbcore import serializers as wb_serializers
from wbcore.content_type.serializers import (
    ContentTypeRepresentationSerializer,
    DynamicObjectIDRepresentationSerializer,
)
from wbcore.contrib.authentication.serializers import (
    GroupRepresentationSerializer,
    PermissionRepresentationSerializer,
    UserRepresentationSerializer,
)
from wbcore.contrib.workflow.models import Process, ProcessStep, Step
from wbcore.contrib.workflow.serializers.step import StepRepresentationSerializer
from wbcore.contrib.workflow.serializers.workflow import (
    WorkflowRepresentationSerializer,
)
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt


class ProcessModelSerializer(wb_serializers.ModelSerializer):
    _instance_id = DynamicObjectIDRepresentationSerializer(
        source="instance_id",
        optional_get_parameters={"content_type": "content_type"},
        depends_on=[{"field": "content_type", "options": {}}],
    )
    _workflow = WorkflowRepresentationSerializer(source="workflow")
    _content_type = ContentTypeRepresentationSerializer(source="content_type")

    @wb_serializers.register_resource()
    def process_steps_inline(self, instance: Process, request, user) -> dict[str, str]:
        if not self.context.get("request"):
            return {}
        return {
            "process_steps": reverse(
                "wbcore:workflow:processstep-process-list",
                args=[instance.pk],
                request=self.context["request"],
            )
        }

    class Meta:
        model = Process
        dependency_map = {
            "instance_id": ["content_type"],
        }
        fields = (
            "id",
            "started",
            "finished",
            "instance_id",
            "_instance_id",
            "workflow",
            "_workflow",
            "_content_type",
            "content_type",
            "_additional_resources",
            "preserved_instance",
            "state",
        )


class ProcessRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:workflow:process-detail")

    class Meta:
        model = Process
        fields = (
            "id",
            "_detail",
        )


class ProcessStepModelSerializer(wb_serializers.ModelSerializer):
    _process = ProcessRepresentationSerializer(source="process")
    _step = StepRepresentationSerializer(source="step")
    _assignee = UserRepresentationSerializer(source="assignee")
    _group = GroupRepresentationSerializer(source="group")
    _permission = PermissionRepresentationSerializer(source="permission")

    @wb_serializers.register_dynamic_button()
    def next_process_step_buttons(self, instance: ProcessStep, request, user):
        buttons = []
        if (
            instance.state == ProcessStep.StepState.ACTIVE
            and instance.step.step_type == Step.StepType.USERSTEP
            and (
                user.is_superuser
                or user == instance.assignee
                or (user in instance.group.user_set.all() and not instance.step.get_casted_step().assignee_method)
            )
        ):
            for transition in instance.step.get_all_valid_outgoing_transitions(instance):
                buttons.append(
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("workflow:step",),
                        endpoint=f"{reverse('wbcore:workflow:processstep-next', args=[instance.pk], request=request)}?transition_id={transition.pk}",
                        label=transition.name,
                        icon=transition.icon,
                        description_fields=_("Are you sure you want to activate {}?").format(transition.name),
                        title=transition.name,
                        action_label=_("Activating {}").format(transition.name),
                    )
                )
        return buttons

    class Meta:
        model = ProcessStep
        fields = (
            "id",
            "started",
            "finished",
            "permission",
            "_permission",
            "process",
            "_process",
            "_step",
            "step",
            "assignee",
            "_assignee",
            "group",
            "_group",
            "state",
            "status",
            "error_message",
            "_additional_resources",
            "_buttons",
        )


class ProcessStepRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcore:workflow:processstep-detail")

    class Meta:
        model = ProcessStep
        fields = (
            "id",
            "_detail",
        )


class AssignedProcessStepSerializer(wb_serializers.ModelSerializer):
    workflow_name = wb_serializers.CharField(label=_("Workflow"))
    _step = StepRepresentationSerializer(source="step")
    _assignee = UserRepresentationSerializer(source="assignee")
    _group = GroupRepresentationSerializer(source="group")
    _permission = PermissionRepresentationSerializer(source="permission")
    instance_endpoint = wb_serializers.SerializerMethodField()
    attached_model = wb_serializers.CharField(label=_("Attached Model"))

    def get_instance_endpoint(self, obj: ProcessStep) -> str:
        """Returns instance endpoint for attached instance object if possible"""

        if not self.context.get("request"):
            return ""
        instance_object = obj.process.instance if obj.process.instance else obj
        return reverse(
            f"{instance_object.get_endpoint_basename()}-detail",
            args=[instance_object.pk],
            request=self.context["request"],
        )

    class Meta:
        model = ProcessStep
        fields = (
            "id",
            "started",
            "finished",
            "permission",
            "_permission",
            "_step",
            "step",
            "assignee",
            "_assignee",
            "group",
            "_group",
            "status",
            "instance_endpoint",
            "workflow_name",
            "attached_model",
        )



from django.contrib import admin
from guardian.admin import GuardedModelAdmin

from wbcore.admin import NonrelatedTabularInline
from wbcore.contrib.documents.models import (
    Document,
    DocumentModelRelationship,
    DocumentType,
    ShareableLink,
    ShareableLinkAccess,
)


class DocumentModelRelationshipInline(admin.TabularInline):
    model = DocumentModelRelationship
    extra = 0
    readonly_fields = ("content_type", "object_id")
    raw_id_fields = ["content_type", "document"]


class ShareableLinkInline(admin.TabularInline):
    model = ShareableLink
    fields = ("valid_until", "one_time_link", "manual_invalid", "uuid")
    readonly_fields = ("uuid",)
    extra = 0
    raw_id_fields = ["document"]
    show_change_link = True


class ShareableLinkAccessInline(admin.TabularInline):
    model = ShareableLinkAccess
    extra = 0
    readonly_fields = [f.name for f in ShareableLinkAccess._meta.get_fields()]
    can_delete = False


class DocumentInLine(NonrelatedTabularInline):
    model = Document
    extra = 0
    fields = ["name", "description", "document_type", "file", "created"]
    readonly_fields = ["created"]
    show_change_link = True

    def get_form_queryset(self, obj):
        return Document.get_for_object(obj)

    def save_new_instance(self, parent, instance):
        instance.save()
        instance.link(parent)


@admin.register(Document)
class DocumentModelAdmin(GuardedModelAdmin):
    inlines = [DocumentModelRelationshipInline, ShareableLinkInline]
    list_display = ["id", "name", "file", "document_type", "created", "updated"]
    search_fields = ["name"]
    raw_id_fields = ["creator"]

    autocomplete_fields = ["creator"]

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("document_type")


@admin.register(DocumentType)
class DocumentTypeModelAdmin(admin.ModelAdmin):
    pass


@admin.register(DocumentModelRelationship)
class DocumentModelRelationshipModelAdmin(admin.ModelAdmin):
    pass


@admin.register(ShareableLink)
class ShareableLinkModelAdmin(admin.ModelAdmin):
    readonly_fields = ["uuid", "document"]
    fieldsets = ((None, {"fields": (("document", "uuid"), ("valid_until", "one_time_link", "manual_invalid"))}),)
    inlines = [ShareableLinkAccessInline]
    list_display = ["uuid", "document", "valid_until", "one_time_link", "manual_invalid"]



from django.contrib.auth import views as auth_views
from django.urls import include, path, reverse_lazy
from rest_framework_simplejwt.views import (  # TokenObtainPairView,; TokenRefreshView,
    TokenVerifyView,
)

from wbcore.routers import WBCoreRouter

from .viewsets import (
    GroupRepresentationViewSet,
    GroupViewSet,
    MyTokenObtainPairModelView,
    MyTokenRefreshModelView,
    PermissionRepresentationViewSet,
    PermissionViewSet,
    UserActivityChart,
    UserActivityModelViewSet,
    UserActivityTable,
    UserActivityUserChart,
    UserActivityUserModelViewSet,
    UserModelViewSet,
    UserPermissionsModelViewSet,
    UserProfileModelViewSet,
    UserRepresentationViewSet,
)
from .viewsets.users import activate_user, register_user, reset_password_email

router = WBCoreRouter()
router.register(r"userrepresentation", UserRepresentationViewSet, basename="userrepresentation")
router.register(r"grouprepresentation", GroupRepresentationViewSet, basename="grouprepresentation")
router.register(r"permissionrepresentation", PermissionRepresentationViewSet, basename="permissionrepresentation")

router.register(r"group", GroupViewSet)
router.register(r"permission", PermissionViewSet)
router.register(r"user", UserModelViewSet, basename="user")
router.register(r"userprofile", UserProfileModelViewSet, basename="userprofile")

router.register(r"useractivitychart", UserActivityChart, basename="useractivitychart")
router.register(r"useractivity", UserActivityModelViewSet, basename="useractivity")
router.register(r"useractivitytable", UserActivityTable, basename="useractivitytable")

user_router = WBCoreRouter()
user_router.register(r"useractivity", UserActivityUserModelViewSet, basename="user-useractivity")
user_router.register(r"useractivitychart", UserActivityUserChart, basename="user-useractivitychart")
user_router.register(
    r"permissions",
    UserPermissionsModelViewSet,
    basename="user-permissions",
)

urlpatterns = [
    # path("token/", TokenObtainPairView.as_view(), name="token_obtain_pair"),
    # path("token/refresh/", TokenRefreshView.as_view(), name="token_refresh"),
    # path("token/verify/", TokenVerifyView.as_view(), name="token_verify"),
    path("token/", MyTokenObtainPairModelView.as_view(), name="token_obtain_pair"),
    path("token/refresh/", MyTokenRefreshModelView.as_view(), name="token_refresh"),
    path("token/verify/", TokenVerifyView.as_view(), name="token_verify"),
    # url(r'^api_token/', get_api_token, name='api-token'),
    # url(r'^wb_menu/', WBMenuAPIView.as_view(), name='wb-menu'),
    path("", include(router.urls)),
    path("user/<int:user_id>/", include(user_router.urls)),
    path(
        "reset_password/",
        auth_views.PasswordResetView.as_view(
            template_name="reset_password.html",
            email_template_name="password_reset_email.html",
            html_email_template_name="password_reset_email_html.html",
            success_url=reverse_lazy("wbcore:authentication:password_reset_done"),
        ),
        name="reset_password",
    ),
    # OR direct Post request view
    path("reset_password_email/", reset_password_email, name="reset_password_email"),
    path(
        "reset_password_sent/",
        auth_views.PasswordResetDoneView.as_view(template_name="password_reset_sent.html"),
        name="password_reset_done",
    ),
    path(
        "reset/<uidb64>/<token>",
        auth_views.PasswordResetConfirmView.as_view(
            template_name="password_reset_form.html",
            success_url=reverse_lazy("wbcore:authentication:password_reset_complete"),
        ),
        name="password_reset_confirm",
    ),
    path(
        "reset_password_complete/",
        auth_views.PasswordResetCompleteView.as_view(template_name="password_reset_done.html"),
        name="password_reset_complete",
    ),
    path("register/", register_user, name="register"),
    path("activate/<uuid>/<token>", activate_user, name="activate"),
]



from django.contrib.contenttypes.models import ContentType
from django.utils.translation import gettext as _

from wbcore.metadata.configs.titles import TitleViewConfig


def get_object_by_content(content_type, content_id):
    return str(ContentType.objects.get_for_id(content_type).model_class().objects.get(id=content_id))


class DocumentModelTitleConfig(TitleViewConfig):
    def get_list_title(self):
        if (content_id := self.view.kwargs.get("content_id", None)) and (
            content_type := self.view.kwargs.get("content_type", None)
        ):
            return _("Documents for {object}").format(object=get_object_by_content(content_type, content_id))
        return _("Documents")

    def get_instance_title(self):
        return _("Document: {{name}}")

    def get_create_title(self):
        if (content_id := self.view.kwargs.get("content_id", None)) and (
            content_type := self.view.kwargs.get("content_type", None)
        ):
            return _("New Document for {object}").format(object=get_object_by_content(content_type, content_id))
        return _("New Document")



import re

from django.contrib import admin
from django.utils.translation import gettext_lazy as _
from reversion.admin import VersionAdmin

from wbcore.admin import ExportCsvMixin, ImportCsvMixin

from ..models import (
    AddressContact,
    BankingContact,
    Company,
    CompanyType,
    CustomerStatus,
    EmailContact,
    Entry,
    Person,
    SocialMediaContact,
    Specialization,
    TelephoneContact,
    WebsiteContact,
)


@admin.register(CustomerStatus)
class CustomerStatusAdmin(admin.ModelAdmin):
    search_fields = ("title",)
    list_display = ("id", "title")


@admin.register(CompanyType)
class CompanyTypeAdmin(admin.ModelAdmin):
    search_fields = ("title",)
    list_display = ("id", "title")


@admin.register(Specialization)
class SpecializationAdmin(admin.ModelAdmin):
    search_fields = ("title",)
    list_display = ("id", "title")


class AddressContactInline(admin.StackedInline):
    model = AddressContact
    raw_id_fields = ["entry", "geography_city"]
    extra = 0


class BankingContactInline(admin.StackedInline):
    model = BankingContact
    extra = 0
    raw_id_fields = ["entry", "currency"]
    autocomplete_fields = ["currency"]


class EmailContactInline(admin.StackedInline):
    model = EmailContact
    extra = 0
    raw_id_fields = ["entry"]


class TelephoneContactInline(admin.StackedInline):
    model = TelephoneContact
    extra = 0
    raw_id_fields = ["entry"]


class WebsiteContactInline(admin.StackedInline):
    model = WebsiteContact
    extra = 0
    raw_id_fields = ["entry"]


class SocialMediaContactInline(admin.StackedInline):
    model = SocialMediaContact
    extra = 0
    raw_id_fields = ["entry"]


class EmployeesInline(admin.StackedInline):
    model = Person
    extra = 0
    raw_id_fields = ["entry", "employers", "specializations"]


@admin.register(Entry)
class EntryAdmin(ExportCsvMixin, ImportCsvMixin, VersionAdmin):
    list_display = ("computed_str", "entry_type", "is_active")
    fieldsets = ((_("Main information"), {"fields": ("computed_str", "is_active")}),)
    search_fields = ("computed_str",)
    raw_id_fields = ("relationship_managers",)
    inlines = [
        EmployeesInline,
        AddressContactInline,
        BankingContactInline,
        EmailContactInline,
        TelephoneContactInline,
        WebsiteContactInline,
        SocialMediaContactInline,
    ]

    def reversion_register(self, model, **options):
        options = {}
        if issubclass(model, Entry):
            options = {
                "follow": (
                    "relationship_managers",
                    "relationships",
                )
            }

        super().reversion_register(model, **options)

    def get_queryset(self, request):
        return self.model.all_objects.all()


@admin.register(Company)
class CompanyAdmin(EntryAdmin):
    fieldsets = (
        (_("Main information"), {"fields": ("name", "computed_str", "slugify_computed_str", "is_active")}),
        (
            _("Additional Fields"),
            {"fields": ("headcount", "tier", "activity_heat", "salutation")},
        ),
        (_("System-Files"), {"fields": ("profile_image",)}),
    )

    def reversion_register(self, model, **options):
        super().reversion_register(model, **options)

    def process_model(self, model):
        managers = []
        if model.get("relationship_managers", None):
            managers = [Person.objects.get(id=person) for person in model.get("relationship_managers").split(",")]
            del model["relationship_managers"]

        obj = self.model.objects.create(**model)
        if managers:
            obj.relationship_managers.add(*managers)


@admin.register(Person)
class PersonAdmin(EntryAdmin):
    search_fields = ("first_name", "last_name")
    list_display = ("first_name", "last_name", "is_active")
    fieldsets = (
        (
            _("Main information"),
            {"fields": ("first_name", "last_name", "computed_str", "slugify_computed_str", "is_active")},
        ),
        (_("Work Fields"), {"fields": ("active_employee",)}),
        (
            _("Additional Fields"),
            {"fields": ("additional_fields",)},
        ),
        (_("Misc."), {"fields": ("birthday",)}),
        (
            _("System-Files"),
            {
                "fields": (
                    "profile_image",
                    "signature",
                )
            },
        ),
        (_("Activity Heat"), {"fields": ("activity_heat",)}),
    )

    def _get_base_fields(self):
        return [f.name for f in self.model._meta.get_fields()]

    def get_import_fields(self):
        emails_fields = [f"email__{e.name}" for e in EmailContact._meta.get_fields()]
        telephone_fields = [f"telephone__{t.name}" for t in TelephoneContact._meta.get_fields()]

        return emails_fields + telephone_fields + self._get_base_fields()

    def reversion_register(self, model, **options):
        options = {}
        if issubclass(self.model, Entry):
            options = {
                "follow": (
                    "employers",
                    "objects",
                    "employees",
                )
            }
        super().reversion_register(model, **options)

    def process_model(self, model):
        employer = relationship_managers = None

        # Extract reverse relation ship (i.e. person's contacts)
        email = {}
        delete_keys = set()
        for k in model.keys():
            m = re.search("email__(.+)", k)
            if m:
                email[m.group(1)] = model[k]
                delete_keys.add(k)

        telephone = {}
        for k in model.keys():
            m = re.search("telephone__(.+)", k)
            if m:
                telephone[m.group(1)] = model[k]
                delete_keys.add(k)
        for key in delete_keys:
            del model[key]
        # Sanitize Foreign keys
        if model.get("employers", None):
            employer, created = Company.objects.get_or_create(name=model["employers"])
            del model["employers"]
        if model.get("relationship_managers", None):
            relationship_managers = Person.objects.get(id=model["relationship_managers"])
            del model["relationship_managers"]

        # Filter non-model fields
        model = dict(filter(lambda elem: elem[0] in self._get_base_fields(), model.items()))
        qs = self.model.objects.filter(first_name=model["first_name"], last_name=model["last_name"])
        if qs.exists():
            obj = qs.first()
        else:
            obj = self.model.objects.create(**model)

        # Create reverse relationship
        if employer:
            obj.employers.add(employer)
        if relationship_managers:
            obj.relationship_managers.add(relationship_managers)
        if telephone and telephone.get("number", None):
            TelephoneContact.objects.get_or_create(entry=obj, number=telephone["number"], defaults=telephone)
        if email and email.get("address", None):
            EmailContact.objects.get_or_create(entry=obj, address=email["address"], defaults=email)



from wbcore import viewsets

from ..filters import TagFilterSet
from ..models import Tag, TagGroup
from ..serializers import (
    TagGroupModelSerializer,
    TagGroupRepresentationSerializer,
    TagModelSerializer,
    TagRepresentationSerializer,
)
from .display import TagDisplayConfig, TagGroupDisplayConfig, TagPreviewConfig


class TagGroupRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = TagGroup.objects.all()
    serializer_class = TagGroupRepresentationSerializer

    search_fields = ["title"]


class TagGroupModelViewSet(viewsets.ModelViewSet):
    queryset = TagGroup.objects.all()
    serializer_class = TagGroupModelSerializer

    filterset_fields = {"title": ["icontains", "exact"]}
    search_fields = ["title"]

    display_config_class = TagGroupDisplayConfig


class TagRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = Tag.objects.all()
    serializer_class = TagRepresentationSerializer
    filterset_class = TagFilterSet

    search_fields = ["title", "slug", "groups__title"]


class TagModelViewSet(viewsets.ModelViewSet):
    queryset = Tag.objects.all()
    serializer_class = TagModelSerializer

    filterset_class = TagFilterSet
    search_fields = ("title", "description")
    ordering_fields = [
        "title",
        "color",
        "slug",
        "content_type",
    ]
    ordering = ["title"]
    display_config_class = TagDisplayConfig
    preview_config_class = TagPreviewConfig


class TagTagGroupModelViewSet(TagModelViewSet):
    def get_queryset(self):
        return super().get_queryset().filter(groups=self.kwargs["group_id"])



# Generated by Django 5.0.8 on 2024-09-09 15:14

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcore", "0013_delete_colorgradient"),
    ]

    operations = [
        migrations.AddField(
            model_name="biguserobjectpermission",
            name="system",
            field=models.BooleanField(default=False),
        ),
        migrations.SeparateDatabaseAndState(
            database_operations=[],
            state_operations=[
                migrations.AlterUniqueTogether(
                    name="biguserobjectpermission",
                    unique_together=None,
                ),
                migrations.RemoveField(
                    model_name="biguserobjectpermission",
                    name="content_type",
                ),
                migrations.RemoveField(
                    model_name="biguserobjectpermission",
                    name="permission",
                ),
                migrations.RemoveField(
                    model_name="biguserobjectpermission",
                    name="user",
                ),
                migrations.DeleteModel(
                    name="BigGroupObjectPermission",
                ),
                migrations.DeleteModel(
                    name="BigUserObjectPermission",
                ),
            ],
        ),
    ]



from pytest_factoryboy import register
from wbcore.contrib.directory.factories import PersonFactory
from wbcore.tests.conftest import *

from ..factories import (
    InternalUserFactory,
    SuperUserFactory,
    TokenFactory,
    UserActivityFactory,
    UserFactory,
)

register(UserFactory, _name="user")
register(SuperUserFactory, _name="superuser")
register(InternalUserFactory, _name="internal_user")
register(UserActivityFactory)
register(PersonFactory)
register(TokenFactory)
